"""
Suite of tests to assess "face validity" of spectral analysis functions in spectra.py
Usually used to test new or majorly updated functions.

Includes tests that parametrically estimate power as a function of frequency, amplitude, phase,
n, etc. to establish methods produce expected pattern of results.

Plots results and runs assertions that basic expected results are reproduced

Functions
---------
- test_power :            Contains tests of spectral estimation functions
- power_test_battery :    Runs standard battery of tests of spectral estimation functions
- itpc_test_battery :     Runs standard battery of tests of ITPC estimation functions
"""
import os
import time
from warnings import warn
from math import pi, sqrt, ceil, floor, log2
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import bernoulli

from spynal.spectra.spectra import power_spectrogram, itpc
from spynal.spectra.utils import simulate_oscillation


def test_power(method, test='frequency', test_values=None, spec_type='power',
               do_tests=True, do_plots=False, plot_dir=None, seed=1,
               amp=5.0, freq=32, phi=0, phi_sd=0, noise=0.5, n=1000, burst_rate=0,
               time_range=3.0, smp_rate=1000, spikes=False, **kwargs):
    """
    Basic testing for functions estimating time-frequency spectral power

    Generate synthetic LFP data using given network simulation,
    estimate spectrogram using given function, and compares estimated to expected.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.
    
    Parameters
    ----------
    method : str
        Name of time-frequency spectral estimation function to test. Options:
        'wavelet' | 'multitaper' | 'bandfilter' | 'burst'

    test : str, default: 'frequency'
        Type of test to run. Options:
        
        - 'frequency' : Tests multiple simulated oscillatory frequencies
            Checks power for monotonic increase of peak freq
        - 'amplitude' : Tests multiple simulated amplitudes at same freq
            Checks power for monotonic increase of amplitude
        - 'phase_sd': Tests multiple simulated phase std dev's (ie evoked vs induced)
            Checks that power doesn't greatly vary with phase SD.
            Checks that ITPC decreases monotonically with phase SD
        - 'n' : Tests multiple values of number of trials (n)
            Checks that power doesn't greatly vary with n.
        - 'burst_rate' : Checks that oscillatory burst rate increases
            as it's increased in simulated data.

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:
        
        - 'frequency' : List of frequencies to test. Default: [4,8,16,32,64]
        - 'amplitude' : List of oscillation amplitudes to test. Default: [1,2,5,10,20]
        - 'n' :         Trial numbers. Default: [25,50,100,200,400,800]

    spec_type : str, default: 'power'
        Type of spectral signal to return. Options: 'power' | 'itpc' (intertrial phase clustering)

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.
        
    - Following args set param's for sim, may be overridden by <test_values> depending on test -
    amp     Scalar. Simulated oscillation amplitude (a.u.) if test != 'amplitude'. Default: 5.0
    freq    Scalar. Simulated oscillation frequency (Hz) if test != 'frequency'. Default: 32
    phi     Scalar. Simulated oscillation (mean) phase (rad). Default: 0
    phi_sd  Scalar. Simulated oscillation phase std dev (rad). Default: 0
    noise   Scalar. Additive noise for simulated signal (a.u., same as amp). Default: 0.5
    n       Int. Number of trials to simulate if test != 'n'. Default: 1000
    burst_rate Scalar. Oscillatory burst rate (bursts/trial). Default: 0 (non-bursty)
    time_range Scalar. Full time range to simulate oscillation over (s). Default: 1.0
    smp_rate Int. Sampling rate for simulated data (Hz). Default: 1000

    **kwargs :
        All other keyword args passed to spectral estimation function

    Returns
    -------
    means : ndarray, shape=(n_freqs,n_timepts,n_values)
        Estimated mean spectrogram for each tested value.

    sems : ndarray, shape=(n_freqs,n_timepts,n_values)
        SEM of mean spectrogram for each tested value.

    passed : bool
        True if all tests produce expected values; otherwise False.
    """
    method = method.lower()
    test = test.lower()
    spec_type = spec_type.lower()

    # Set defaults for tested values and set up rate generator function depending on <test>
    sim_args = dict(amplitude=amp, phase=phi, phase_sd=phi_sd,
                    n_trials=n, noise=noise, time_range=time_range, burst_rate=burst_rate,
                    seed=seed)

    if test in ['frequency','freq']:
        test_values = [4,8,16,32,64] if test_values is None else test_values
        gen_data = lambda freq: simulate_oscillation(freq,**sim_args)

    elif test in ['amplitude','amp']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['amplitude']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda amp: simulate_oscillation(freq,**sim_args,amplitude=amp)

    elif test in ['phase','phi']:
        test_values = [-pi,-pi/2,0,pi/2,pi] if test_values is None else test_values
        del sim_args['phase']       # Delete preset arg so it uses argument to lambda below
        gen_data = lambda phi: simulate_oscillation(freq,**sim_args,phase=phi)

    elif test in ['phase_sd','phi_sd']:
        test_values = [pi, pi/2, pi/4, 0] if test_values is None else test_values
        del sim_args['phase_sd']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda phi_sd: simulate_oscillation(freq,**sim_args,phase_sd=phi_sd)

    elif test in ['n','n_trials']:
        test_values = [25,50,100,200,400,800] if test_values is None else test_values
        del sim_args['n_trials']    # Delete preset arg so it uses argument to lambda below
        gen_data = lambda n: simulate_oscillation(freq,**sim_args,n_trials=n)

    elif test in ['burst_rate','burst']:
        test_values = [0.1,0.2,0.4,0.8] if test_values is None else test_values
        del sim_args['burst_rate']  # Delete preset arg so it uses argument to lambda below
        gen_data = lambda rate: simulate_oscillation(freq,**sim_args,burst_rate=rate)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    # Ensure hand-set values are sorted (ascending), as many tests assume it
    test_values = sorted(test_values)
    n_values = len(test_values)

    # Set default parameters for each spectral estimation method
    do_burst = method in ['burst','burst_analysis']
    do_itpc = spec_type == 'itpc'

    # Special case: oscillatory burst analysis
    if do_burst:
        # KLUDGE  Reset spectral analysis <method> to 'wavelet' (unless set explicitly in kwargs)
        if 'bands' not in kwargs:       kwargs['bands'] = ((2,6),(6,10),(10,22),(22,42),(42,86))

    elif method == 'multitaper':
        if 'freq_range' not in kwargs:  kwargs['freq_range'] = [1,100]

    elif method == 'bandfilter':
        if 'freqs' not in kwargs:       kwargs['freqs'] = ((2,6),(6,10),(10,22),(22,42),(42,86))

    if ('buffer' not in kwargs) and (method != 'multitaper'): kwargs['buffer'] = 1.0

    if do_itpc:
        if 'itpc_method' not in kwargs: kwargs['itpc_method'] = 'PLV'
        kwargs['trial_axis'] = 1

    spec_fun = itpc if do_itpc else power_spectrogram

    for i,value in enumerate(test_values):
        # print("Running test value %d/%d: %.2f" % (i+1,n_values,value))

        # Simulate data with oscillation of given params -> (n_timepts,n_trials)
        data = gen_data(value)

        # HACK Convert continuous oscillatory data into spike train (todo find better method)
        if spikes:
            data = (data - data.min()) / data.ptp() # Convert to 0-1 range ~ spike probability
            data = data**2                          # Sparsify probabilies (decrease rates)
            # Use probabilities to generate Bernoulli random variable at each time point
            data = bernoulli.ppf(0.5, data).astype(bool)

        spec,freqs,timepts = spec_fun(data,smp_rate,axis=0,method=method,**kwargs)
        if freqs.ndim == 2:
            bands = freqs
            freqs = freqs.mean(axis=1)  # Compute center of freq bands
        if do_itpc: n_freqs,n_timepts = spec.shape
        else:       n_freqs,n_timepts,n_trials = spec.shape

        # KLUDGE Initialize output arrays on 1st loop, once spectrogram output shape is known
        if i == 0:
            means = np.empty((n_freqs,n_timepts,n_values))
            sems = np.empty((n_freqs,n_timepts,n_values))

        # Compute across-trial mean and SEM of time-frequency data -> (n_freqs,n_timepts,n_values)
        if not do_itpc:
            means[:,:,i] = spec.mean(axis=2)
            sems[:,:,i]  = spec.std(axis=2,ddof=0) / sqrt(n_trials)
        # HACK ITPC by definition already reduced across trials, so just copy results into "means"
        else:
            means[:,:,i] = spec
            sems[:,:,i]  = 0

    # Compute mean across all timepoints -> (n_freqs,n_values) frequency marginal
    marginal_means = means.mean(axis=1)
    marginal_sems = sems.mean(axis=1)

    # For bandfilter, plot frequency bands in categorical fashion
    if do_burst or (method == 'bandfilter'):
        freq_transform  = lambda x: np.argmin(np.abs(x - freqs))  # Index of closest sampled freq
        plot_freqs      = np.arange(n_freqs)
        freq_ticks      = np.arange(n_freqs)
        freq_tick_labels= bands

    # For wavelets, evaluate and plot frequency on log scale
    elif method == 'wavelet':
        freq_transform  = np.log2
        plot_freqs      = freq_transform(freqs)
        fmin            = ceil(log2(freqs[0]))
        fmax            = floor(log2(freqs[-1]))
        freq_ticks      = np.arange(fmin,fmax+1)
        freq_tick_labels= 2**np.arange(fmin,fmax+1)

    # For multitaper, evaluate and plot frequency on linear scale
    elif method == 'multitaper':
        freq_transform  = lambda x: x
        plot_freqs      = freqs
        fmin            = ceil(freqs[0]/10.0)*10.0
        fmax            = floor(freqs[-1]/10.0)*10.0
        freq_ticks      = np.arange(fmin,fmax+1,10).astype(int)
        freq_tick_labels= freq_ticks

    freqs_transformed   = np.asarray([freq_transform(f) for f in freqs])

    # For frequency test, find frequency with maximal power for each test
    if test in ['frequency','freq']:
        idxs = np.argmax(marginal_means,axis=0)
        peak_freqs = freqs[idxs] if not(do_burst or (method == 'bandfilter')) else idxs

        # Find frequency in spectrogram closest to each simulated frequency
        test_freq_idxs  = np.asarray([np.argmin(np.abs(freq_transform(f) - freqs_transformed))
                                      for f in test_values])

        # Extract mean,SEM of power at each tested frequency
        test_freq_means = marginal_means[test_freq_idxs,np.arange(n_values)]
        test_freq_errs  = marginal_sems[test_freq_idxs,np.arange(n_values)]

    else:
        # Find frequency in spectrogram closest to simulated frequency
        test_freq_idx   = np.argmin(np.abs(freq_transform(freq) - freqs_transformed))

        # Extract mean,SEM of power at tested frequency
        test_freq_means = marginal_means[test_freq_idx,:]
        test_freq_errs  = marginal_sems[test_freq_idx,:]

    # Plot summary of test results
    if do_plots:
        dt      = np.diff(timepts).mean()
        tlim    = [timepts[0]-dt/2, timepts[-1]+dt/2]
        df      = np.diff(plot_freqs).mean()
        flim    = [plot_freqs[0]-df/2, plot_freqs[-1]+df/2]

        # # Plot spectrogram for each tested value
        # plt.figure()
        # n_subplots = [floor(n_values/2), ceil(n_values/floor(n_values/2))]
        # for i,value in enumerate(test_values):
        #     ax = plt.subplot(n_subplots[0],n_subplots[1],i+1)
        #     plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        #     target_freq = freq_transform(value) if test in ['frequency','freq'] else \
        #                   freq_transform(freq)
        #     if not (do_burst or (method == 'bandfilter')):
        #         plt.plot(tlim, [target_freq,target_freq], '-', color='r', linewidth=0.5)
        #     plt.imshow(means[:,:,i], extent=[*tlim,*flim], aspect='auto', origin='lower')
        #     if i in [0,n_subplots[1]]:
        #         plt.yticks(freq_ticks,freq_tick_labels)
        #     else:
        #         ax.set_xticklabels([])
        #         plt.yticks(freq_ticks,[])
        #     plt.title(np.round(value,decimals=2))
        #     plt.colorbar()
        # plt.show()
        # if plot_dir is not None:
        #     filename = 'power-spectrogram-%s-%s-%s.png' % (kwargs['itpc_method'],method,test) \
        #                 if do_itpc else 'power-spectrogram-%s-%s.png' % (method,test)
        #     plt.savefig(os.path.join(plot_dir,filename))

        # Plot time-averaged spectrum for each tested value
        plt.figure()
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        ylim = [0,1.05*marginal_means.max()]
        for i,value in enumerate(test_values):
            plt.plot(plot_freqs, marginal_means[:,i], '.-', color=colors[i], linewidth=1.5)
            target_freq = freq_transform(value) if test in ['frequency','freq'] else \
                          freq_transform(freq)
            if not (do_burst or (method == 'bandfilter')):
                plt.plot([target_freq,target_freq], ylim, '-', color=colors[i], linewidth=0.5)
            plt.text(0.9*flim[1], (0.95-i*0.05)*ylim[1], value, color=colors[i], fontweight='bold')
        plt.xlim(flim)
        plt.ylim(ylim)
        plt.xticks(freq_ticks,freq_tick_labels)
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.xlabel('Frequency (Hz)')
        plt.ylabel(spec_type)
        plt.title("%s %s test" % (method,test))
        plt.show()
        if plot_dir is not None:
            filename = 'power-spectrum-%s-%s-%s.png' % (kwargs['itpc_method'],method,test) \
                        if do_itpc else 'power-spectrum-%s-%s.png' % (method,test)
            plt.savefig(os.path.join(plot_dir,filename))

        # Plot summary curve of power (or peak frequency) vs tested value
        plt.figure()
        ax = plt.subplot(1,1,1)
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        if test in ['frequency','freq']:
            lim = (0,1.1*freq_transform(test_values[-1]))
            plt.plot(lim, lim, color='k', linewidth=0.5)
            if do_burst or (method == 'bandfilter'):
                plt.plot([freq_transform(f) for f in test_values], peak_freqs, marker='o')
            else:
                plt.plot([freq_transform(f) for f in test_values],
                         [freq_transform(f) for f in peak_freqs], marker='o')
            plt.xticks(freq_ticks,freq_tick_labels)
            plt.yticks(freq_ticks,freq_tick_labels)
            plt.xlim(lim)
            plt.ylim(lim)
            ax.set_aspect('equal', 'box')
        else:
            plt.errorbar(test_values, test_freq_means, 3*test_freq_errs, marker='o')
        plt.xlabel(test)
        plt.ylabel('frequency' if test in ['frequency','freq'] else spec_type)
        plt.title("%s %s test" % (method,test))
        plt.show()
        if plot_dir is not None:
            filename = 'power-summary-%s-%s-%s.png' % (kwargs['itpc_method'],method,test) \
                        if do_itpc else 'power-summary-%s-%s.png' % (method,test)
            plt.savefig(os.path.join(plot_dir,filename))

    ## Determine if test actually produced the expected values
    # frequency test: check if frequency of peak power matches simulated target frequency
    if test in ['frequency','freq']:
        evals = [((np.diff(peak_freqs) >= 0).all(),
                    "Estimated peak freq does not increase monotonically with expected freq")]

    # 'amplitude' : Test if power increases monotonically with simulated amplitude
    elif test in ['amplitude','amp']:
        if spec_type == 'power':
            evals = [((np.diff(test_freq_means) > 0).all(),
                        "Estimated power doesn't increase monotonically with simulated amplitude")]
        else:
            evals = {}

    # 'phase' : Test if power is ~ constant across phase
    elif test in ['phase','phi']:
        crit = 0.2 if do_itpc else test_freq_errs.max()
        evals = [(test_freq_means.ptp() < crit,
                    "Estimated %s has larger than expected range across different simulated phases"
                    % spec_type)]

    # 'phase_sd' : Test if power is ~ constant across phase SD;
    #              Test if ITPC decreases monotonically with it
    elif test in ['phase_sd','phi_sd']:
        if do_itpc:
            evals = [((np.diff(test_freq_means) < 0).all(),
                        "Estimated ITPC does not decrease monotonically with simulated phase SD")]
        else:
            evals = [(test_freq_means.ptp() < test_freq_errs.max(),
                        "Estimated %s has larger than expected range across simulated phase SDs"
                        % spec_type)]

    # 'n' : Test if power is ~ same for all values of n (unbiased by n)
    elif test in ['n','n_trials']:
        crit = 0.2 if do_itpc else test_freq_errs.max()
        evals = [(test_freq_means.ptp() < crit,
                    "Estimated %s has larger than expected range across n's (likely biased by n)"
                    % spec_type)]

    # 'burst_rate': Test if measured burst rate increases monotonically with simulated burst rate
    elif test in ['burst_rate','burst']:
        evals = [((np.diff(test_freq_means) > 0).all(),
                    "Estimated burst rate does not increase monotonic with simulated burst rate")]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return means, sems, passed


def power_test_battery(methods=('wavelet','multitaper','bandfilter'),
                       tests=('frequency','amplitude','phase','phase_sd','n','burst_rate'),
                       do_tests=True, **kwargs):
    """
    Run a battery of given tests on given oscillatory power computation methods

    Parameters
    ----------
    methods : array-like of str, default: ('wavelet','multitaper','bandfilter') (all supported)
        List of power computation methods to test.
                
    tests : array-like of str, default: ('frequency','amplitude','phase','phase_sd','n','burst_rate')
        List of tests to run.

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail.

    **kwargs :
        Any other keyword args passed directly to test_power()
    """
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]

    for test in tests:
        for method in methods:
            print("Running %s test on %s spectral analysis" % (test,method))
            extra_args = kwargs
            if (method in ['burst','burst_analysis']) and ('burst_rate' not in kwargs):
                extra_args['burst_rate'] = 0.4

            t1 = time.time()

            _,_,passed = test_power(method, test=test, do_tests=do_tests, **extra_args)
            print('%s (test ran in %.1f s)' % ('PASSED' if passed else 'FAILED', time.time()-t1))

            # If saving plots to file, let's not leave them all open
            if 'plot_dir' in kwargs: plt.close('all')


def itpc_test_battery(methods=('wavelet','multitaper','bandfilter'),
                      tests=('frequency','amplitude','phase','phase_sd','n'),
                      itpc_methods=('PLV','Z','PPC'), do_tests=True, **kwargs):
    """
    Run a battery of given tests on given intertrial phase clustering computation methods

    Parameters
    ----------
    methods : array-like, default: ('wavelet','multitaper','bandfilter') (all supported methods)
        List of power computation methods to test.
                

    tests : array-like, default: ('frequency','amplitude','phase','phase_sd','n') (all supported)
        List of tests to run.                

    itpc_method : array-like, default: ('PLV','Z','PPC') (all supported options)
        List of methods to use for computing intertrial phase clustering
                
    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail.

    **kwargs :
        Any other keyword args passed directly to test_power()
    """
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]

    # Default phase SD = 90 deg unless set otherwise
    phi_sd = kwargs.pop('phi_sd',pi/4)

    for test in tests:
        for itpc_method in itpc_methods:
            for method in methods:
                print("Running %s test on %s %s" % (test,method,itpc_method))
                extra_args = kwargs
                t1 = time.time()

                _,_,passed = test_power(method, test=test, itpc_method=itpc_method,
                                        spec_type='itpc', phi_sd=phi_sd, do_tests=do_tests,
                                        **extra_args)
                print('%s (test ran in %.1f s)'
                      % ('PASSED' if passed else 'FAILED', time.time()-t1))

                # If saving plots to file, let's not leave them all open
                if 'plot_dir' in kwargs: plt.close('all')



# -*- coding: utf-8 -*-
"""
Private helper functions for spynal code

These are functions used internally in one or more analysis modules,
but not intended to be public-facing.
"""
# Created on Fri Apr  9 14:08:15 2021
#
# @author: sbrincat
from copy import deepcopy
import numpy as np

from scipy.stats import mode


def _check_window_lengths(windows, tol=1):
    """
    Ensures a set of windows are the same length. If not equal, but within given tolerance,
    windows are trimmed or expanded to the modal window length.

    Parameters
    ----------
    windows : array-like, shape=(n_wins,2)
        Set of windows to test, given as series of [start,end].

    tol : scalar, default: 1
        Max tolerance of difference of each window length from the modal value.

    Returns
    -------
    windows : ndarray, shape=(n_wins,2)
        Same windows, possibly slightly trimmed/expanded to uniform length
    """
    windows = np.asarray(windows)

    window_lengths  = np.diff(windows,axis=1).squeeze()
    window_range    = np.ptp(window_lengths)

    # If all window lengths are the same, windows are OK and we are done here
    if np.allclose(window_lengths, window_lengths[0]): return windows

    # Compute mode of windows lengths and max difference from it
    modal_length    = mode(window_lengths)[0][0]
    max_diff        = np.max(np.abs(window_lengths - modal_length))

    # If range is beyond our allowed tolerance, throw an error
    assert max_diff <= tol, \
        ValueError("All windows must have same length (input range=%.1f)" % window_range)

    # If range is between 0 and tolerance, we trim/expand windows to the modal length
    windows[:,1]    = windows[:,1] + (modal_length - window_lengths)
    return windows


def _enclose_in_object_array(data):
    """ Enclose array within an object array """
    out = np.empty((1,),dtype=object)
    out[0] = data
    return out


def _isbinary(x):
    """ Test whether variable contains only binary values in set {True,False,0,1} """
    x = np.asarray(x)
    return (x.dtype == bool) or \
           (np.issubdtype(x.dtype,np.number) and \
            np.all(np.in1d(x,[0,0.0,1,1.0,True,False])))


def _has_method(obj, method):
    """
    Determine if given object class instance has given method

    Parameters
    ----------
    obj : object class instance (of any type)
        Object to test for presence of given method

    method : str
        Name of method to test for

    Returns
    -------
    tf : bool
        True if obj.method exists; False otherwise

    References
    ----------
    https://stackoverflow.com/questions/7580532/how-to-check-whether-a-method-exists-in-python/7580687
    """
    return callable(getattr(obj, method, None))


def _merge_dicts(dict1, dict2):
    """ Merge two dictionaries, with values in dict2 overriding (default) values in dict1 """
    dict_out = deepcopy(dict1)
    dict_out.update(dict2)
    return dict_out


def _standardize_to_axis_0(data, axis=0):
    """
    Reshape multi-dimensional data array to standardized 2D array (matrix-like) form,
    with "business" axis shifted to axis 0 for analysis

    Parameters
    ----------
    data : ndarray, shape=(...,n,...). Data array of arbitrary shape.

    axis : int, default: 0
        Axis of data to move to axis 0 for subsequent analysis

    Returns
    -------
    data : ndarray, shape=(n,m)
        Data array w/ `axis` moved to axis=0, and all other axes unwrapped into single dimension,
        where m = prod(shape[axes != axis])

    data_shape : tuple, shape=(data.ndim,)
        Original shape of data array

    NOTE: Even 1d (vector) data is expanded into 2d (n,1) array to standardize for calling code.
    """
    # Save original shape/dimensionality of <data>
    data_ndim  = data.ndim
    data_shape = data.shape

    if ~data.flags.c_contiguous:
        # If observation axis != 0, permute axis to make it so
        if axis != 0:       data = np.moveaxis(data,axis,0)

        # If data array data has > 2 dims, keep axis 0 and unwrap other dims into a matrix
        if data_ndim > 2:   data = np.reshape(data,(data_shape[axis],-1),order='F')

    # Faster method for c-contiguous arrays
    else:
        # If observation axis != last dim, permute axis to make it so
        lastdim = data_ndim - 1
        if axis != lastdim: data = np.moveaxis(data,axis,lastdim)

        # If data array data has > 2 dims, keep axis 0 and unwrap other dims
        # into a matrix, then transpose
        if data_ndim > 2:   data = np.reshape(data,(-1,data_shape[axis]),order='C').T
        else:               data = data.T

    # Expand (n,) 1d data to (n,1) to simplify downstream code
    if data_ndim == 1:  data = data[:,np.newaxis]

    return data, data_shape


def _undo_standardize_to_axis_0(data, data_shape, axis=0):
    """
    Undo effect of _standardize_to_axis_0() -- reshapes data array from unwrapped
    2D (matrix-like) form back to ~ original multi-dimensional form, with `axis`
    shifted back to original location (but allowing that data.shape[axis] may have changed)

    Parameters
    ----------
    data : ndarray, shape=(axis_len,m)
        Data array w/ `axis` moved to axis=0, and all axes != 0 unwrapped into single dimension,
        where m = prod(shape[1:])

    data_shape : tuple, shape=(data_orig.ndim,)
        Original shape of data array. Second output of :func:`_standardize_to_axis_0`.

    axis : int, default: 0
        Axis of original data moved to axis 0, which will be shifted back to original axis.

    Returns
    -------
    data : ndarray, shape=(...,axis_len,...)
        Data array reshaped back to original shape
    """
    data_ndim = len(data_shape) # Number of dimensions in original data
    axis_len  = data.shape[0]   # Length of dim 0 (will become dim <axis> again)

    # If data array data had > 2 dims, reshape matrix back into ~ original shape
    # (but with length of dimension <axis> = <axisLength>)
    if data_ndim > 2:
        # Reshape data -> (axis_len,<original shape w/o <axis>>)
        shape = (axis_len, *data_shape[np.arange(data_ndim) != axis])
        # Note: I think you want the order to be 'F' regardless of memory layout
        data = np.reshape(data,shape,order='F')

    # Squeeze (n,1) array back down to 1d (n,) vector,
    #  and extract value from scalar array -> float
    elif data_ndim == 1:
        data = data.squeeze(axis=-1)
        if data.size == 1: data = data.item()

    # If observation axis wasn't 0, permute axis back to original position
    if (axis != 0) and isinstance(data,np.ndarray):
        data = np.moveaxis(data,0,axis)

    return data


def _standardize_to_axis_end(data, axis=-1):
    """
    Reshape multi-dimensional data array to standardized 2D array (matrix-like) form,
    with "business" axis shifted to axis -1 (end) for analysis

    Parameters
    ----------
    data : ndarray, shape=(...,n,...)
        Data array of arbitrary shape.

    axis : int, default: -1
        Axis of data to move to axis -1 for subsequent analysis. 

    Returns
    -------
    data : ndarray, shape=(m,n)
        Data array w/ `axis` moved to axis=-1, and all other axes unwrapped into single dimension,
        where m = prod(shape[axes != axis])

    data_shape : tuple, shape=(data.ndim,)
        Original shape of data array

    NOTE: Even 1d (vector) data is expanded into 2d (1,n) array to standardize for calling code.
    """
    if axis < 0: axis = data.ndim + axis
    data = np.asarray(data)

    # Save original shape/dimensionality of <data>
    data_ndim  = data.ndim
    data_shape = data.shape

    # Faster method for f-contiguous arrays
    if data.flags.f_contiguous:
        # If observation axis != first dim, permute axis to make it so
        if axis != 0: data = np.moveaxis(data,axis,0)

        # If data array data has > 2 dims, keep axis 0 and unwrap other dims
        # into a matrix, then transpose
        if data_ndim > 2:   data = np.reshape(data,(data_shape[axis],-1),order='F').T
        else:               data = data.T

    else:
        # If observation axis != -1, permute axis to make it so
        if axis != data_ndim - 1: data = np.moveaxis(data,axis,-1)

        # If data array data has > 2 dims, keep axis -1 and unwrap other dims into a matrix
        if data_ndim > 2:   data = np.reshape(data,(-1,data_shape[axis]),order='C')

    # Expand (n,) 1d data to (1,n) to simplify downstream code
    if data_ndim == 1:  data = data[np.newaxis,:]

    return data, data_shape


def _undo_standardize_to_axis_end(data, data_shape, axis=-1):
    """
    Undo effect of :func:`_standardize_to_axis_end` -- reshapes data array from unwrapped
    2D (matrix-like) form back to ~ original multi-dimensional form, with `axis`
    shifted back to original location (but allowing that data.shape[axis] may have changed)

    Parameters
    ----------
    data : ndarray, shape=(m,axis_len)
        Data array w/ <axis> moved to axis=-1, and all axes != -1 unwrapped into single dimension,
        where m = prod(shape[:-1])

    data_shape : tuple, shape=(data_orig.ndim,)
        Original shape of data array. Second output of :func:`_standardize_to_axis_end`.

    axis : int, default: -1
        Axis of original data moved to axis -1, which will be shifted back to original axis.

    Returns
    -------
    data : ndarray, shape=(...,axis_len,...)
        Data array reshaped back to original shape
    """
    data_shape  = np.asarray(data_shape)

    data_ndim   = len(data_shape) # Number of dimensions in original data
    axis_len    = data.shape[-1]  # Length of dim -1 (will become dim <axis> again)

    # If data array data had > 2 dims, reshape matrix back into ~ original shape
    # (but with length of dimension <axis> = <axis_length>)
    if data_ndim > 2:
        # Reshape data -> (<original shape w/o <axis>>,axis_len)
        shape = (*data_shape[np.arange(data_ndim) != axis], axis_len)
        # Note: I think you want the order to be 'C' regardless of memory layout
        data  = np.reshape(data,shape,order='C')

    # Squeeze (1,n) array back down to 1d (n,) vector,
    #  and extract value from scalar array -> float
    elif data_ndim == 1:
        data = data.squeeze(axis=0)
        if data.size == 1: data = data.item()

    # If observation axis wasn't -1, permute axis back to original position
    if (axis != -1) and isinstance(data,np.ndarray):
        data = np.moveaxis(data,-1,axis)

    return data



""" Unit tests for spectra.py module """
from collections import OrderedDict
import pytest
import numpy as np
import xarray as xr

from spynal.utils import iarange
from spynal.tests.data_fixtures import oscillation, bursty_oscillation, spiking_oscillation, \
                                       oscillatory_data, MISSING_ARG_ERRS
from spynal.spectra.spectra import spectrum, spectrogram, itpc, plot_spectrum, plot_spectrogram
from spynal.spectra.preprocess import cut_trials, realign_data, remove_dc, remove_evoked
from spynal.spectra.postprocess import one_over_f_norm, pool_freq_bands, pool_time_epochs
from spynal.spectra.utils import get_freq_sampling, one_sided_to_two_sided, power


# =============================================================================
# Unit tests for spectral analysis functions
# =============================================================================
@pytest.mark.parametrize('data_type, spec_type, method, result',
                         [('lfp',   'power', 'multitaper',  0.0137),
                          ('lfp',   'power', 'wavelet',     44.9462),
                          ('lfp',   'power', 'bandfilter',  2.3069),
                          ('spike', 'power', 'multitaper',  194.0514),
                          ('spike', 'power', 'wavelet',     0.4774),
                          ('spike', 'power', 'bandfilter',  0.0483)])
def test_spectrum(oscillatory_data, data_type, spec_type, method, result):
    """ Unit tests for spectrum function """
    data = oscillatory_data[data_type]
    data_orig = data.copy()
    smp_rate = 1000
    n_trials = 4
    method_to_n_freqs   = {'wavelet': 26, 'multitaper':513,  'bandfilter': 3}
    n_freqs = method_to_n_freqs[method]
    freqs_shape = (n_freqs,2) if (method == 'bandfilter') or (spec_type == 'burst') else (n_freqs,)

    # Time reversal -> inverted sign phase, complex conj of complex, preserves power
    if spec_type == 'phase':        reversed_result = -result
    elif spec_type == 'complex':    reversed_result = np.conj(result)
    else:                           reversed_result = result

    # Basic test of shape, dtype, value of output.
    # Test values averaged over all freqs in first trial for simplicity
    spec, freqs = spectrum(data, smp_rate, axis=0, method=method, data_type=data_type,
                           spec_type=spec_type)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert spec.shape == (n_freqs, n_trials)
    assert np.issubdtype(spec.dtype,float)
    assert np.isclose(spec[:,0].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    spec, freqs = spectrum(data.reshape((-1,int(n_trials/2),int(n_trials/2))),
                           smp_rate, axis=0, method=method, data_type=data_type,
                           spec_type=spec_type)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert spec.shape == (n_freqs, n_trials/2, n_trials/2)
    assert np.issubdtype(spec.dtype,float)
    assert np.isclose(spec[:,0,0].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with transposed data dimensionality
    spec, freqs = spectrum(data.T, smp_rate, axis=-1, method=method, data_type=data_type,
                           spec_type=spec_type)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert spec.shape == (n_trials, n_freqs)
    assert np.issubdtype(spec.dtype,float)
    assert np.isclose(spec[0,:].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with vector-valued data
    spec, freqs = spectrum(data[:,0], smp_rate, axis=-1, method=method, data_type=data_type,
                           spec_type=spec_type)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert spec.shape == (n_freqs,)
    assert np.issubdtype(spec.dtype,float)
    assert np.isclose(spec.mean(), result, rtol=1e-4, atol=1e-4)

    # Test for expected output with time-reversed data
    # Skip test for bandfilter method -- different init conds do change results slightly at start
    # Skip test for multitaper phase/complex -- not time-reversal invariant
    if (method == 'wavelet') or ((method == 'multitaper') and (spec_type == 'power')):
        spec, freqs = spectrum(np.flip(data,axis=0), smp_rate, axis=0, method=method,
                               data_type=data_type, spec_type=spec_type)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert spec.shape == (n_freqs, n_trials)
        assert np.isclose(spec[:,0].mean(), reversed_result, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        spec, freqs = spectrum(data, smp_rate, axis=0, method=method, data_type=data_type,
                            spec_type=spec_type, foo=None)


# TODO Take a closer look at 'complex' outputs -- why coming out = 0?
@pytest.mark.parametrize('data_type, spec_type, method, result',
                         [('lfp',   'power',    'wavelet',     44.9462),
                          ('lfp',   'power',    'multitaper',  0.0137),
                          ('lfp',   'power',    'bandfilter',  2.3069),
                          ('burst', 'burst',    'wavelet',     0.0),
                          ('burst', 'burst',    'bandfilter',  0.0),
                          ('spike', 'power',    'wavelet',     0.4774),
                          ('spike', 'power',    'multitaper',  193.7612),
                          ('spike', 'power',    'bandfilter',  0.0483),
                          ('lfp',   'phase',    'wavelet',     0.0053),
                          ('lfp',   'phase',    'multitaper', -0.1523),
                          ('lfp',   'phase',    'bandfilter',  0.1473),
                          ('spike', 'phase',    'wavelet',     0.0054),
                          ('spike', 'phase',    'multitaper', -0.0547),
                          ('spike', 'phase',    'bandfilter',  0.1190),
                          ('lfp',   'complex',  'wavelet',     0j),
                          ('lfp',   'complex',  'multitaper',  0.0019-0.0011j),
                          ('lfp',   'complex',  'bandfilter',  -0.0109+0j)])
def test_spectrogram(oscillatory_data, data_type, spec_type, method, result):
    """ Unit tests for spectrogram() function """
    # Extract given data type from data dict
    data = oscillatory_data[data_type]
    data_orig = data.copy()
    data_type_ = 'lfp' if data_type == 'burst' else data_type
    smp_rate = 1000
    n_trials = 4

    method_to_n_freqs   = {'wavelet': 26, 'multitaper':257, 'bandfilter': 3, 'burst':4}
    method_to_n_timepts = {'wavelet': 1000, 'multitaper':2, 'bandfilter': 1000}
    n_freqs = method_to_n_freqs['burst'] if spec_type == 'burst' else method_to_n_freqs[method]
    freqs_shape = (n_freqs,2) if (method == 'bandfilter') or (spec_type == 'burst') else (n_freqs,)
    n_timepts = method_to_n_timepts[method]
    if spec_type == 'complex':  dtype = complex
    elif spec_type == 'burst':  dtype = bool
    else:                       dtype = float

    # Time reversal -> inverted sign phase, complex conj of complex, preserves power
    if spec_type == 'phase':        reversed_result = -result
    elif spec_type == 'complex':    reversed_result = np.conj(result)
    else:                           reversed_result = result

    # Basic test of shape, dtype, value of output.
    # Test values averaged over all timepts, freqs for 1st trial for simplicity
    spec, freqs, timepts = spectrogram(data, smp_rate, axis=0, method=method,
                                       data_type=data_type_, spec_type=spec_type)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert spec.shape == (n_freqs, n_timepts, n_trials)
    assert np.issubdtype(spec.dtype,dtype)
    assert np.isclose(spec[:,:,0].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    spec, freqs, timepts = spectrogram(data.reshape((-1,int(n_trials/2),int(n_trials/2))),
                                       smp_rate, axis=0, method=method, data_type=data_type_,
                                       spec_type=spec_type)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert spec.shape == (n_freqs, n_timepts, n_trials/2, n_trials/2)
    assert np.issubdtype(spec.dtype,dtype)
    assert np.isclose(spec[:,:,0,0].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with transposed data dimensionality
    extra_args = {'trial_axis':0} if spec_type == 'burst' else {}
    spec, freqs, timepts = spectrogram(data.T, smp_rate, axis=-1, method=method,
                                       data_type=data_type_, spec_type=spec_type, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert spec.shape == (n_trials, n_freqs, n_timepts)
    assert np.issubdtype(spec.dtype,dtype)
    assert np.isclose(spec[0,:,:].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with vector-valued (single-trial/channel) data
    # Note: Skip for burst_analysis bc needs multiple trials to compute z-scores
    if spec_type != 'burst':
        spec, freqs, timepts = spectrogram(data[:,0], smp_rate, axis=0, method=method,
                                        data_type=data_type_, spec_type=spec_type)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert freqs.shape == freqs_shape
        assert timepts.shape == (n_timepts,)
        assert spec.shape == (n_freqs, n_timepts)
        assert np.issubdtype(spec.dtype,dtype)
        assert np.isclose(spec[:,:].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for expected output with time-reversed data
    # Skip test for bandfilter method -- different init conds do change results slightly at start
    # Skip test for multitaper phase/complex -- not time-reversal invariant
    if (method == 'wavelet') or ((method == 'multitaper') and (spec_type == 'power')):
        spec, freqs, timepts = spectrogram(np.flip(data,axis=0), smp_rate, axis=0, method=method,
                                        data_type=data_type_, spec_type=spec_type)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert spec.shape == (n_freqs, n_timepts, n_trials)
        assert np.isclose(spec[:,:,0].mean(), reversed_result, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        spec, freqs, timepts = spectrogram(data, smp_rate, axis=0, method=method,
                                        data_type=data_type_, spec_type=spec_type, foo=None)


@pytest.mark.parametrize('itpc_method, method, result',
                         [('plv',   'wavelet',     0.5494),
                          ('plv',   'multitaper',  0.2705),
                          ('plv',   'bandfilter',  0.6556),
                          ('z',     'wavelet',     1.5316),
                          ('z',     'multitaper',  0.3762),
                          ('z',     'bandfilter',  2.0734),
                          ('ppc',   'wavelet',     0.1772),
                          ('ppc',   'multitaper',  -0.2079),
                          ('ppc',   'bandfilter',  0.3578)])
def test_itpc(oscillation, itpc_method, method, result):
    """ Unit tests for spectrogram() function """
    # Extract given data type from data dict
    data = oscillation
    data_orig = data.copy()
    smp_rate = 1000
    n_trials = 4

    method_to_n_freqs   = {'wavelet': 26, 'multitaper':257, 'bandfilter': 3, 'burst':4}
    method_to_n_timepts = {'wavelet': 1000, 'multitaper':2, 'bandfilter': 1000}
    n_freqs = method_to_n_freqs[method]
    freqs_shape = (n_freqs,2) if method == 'bandfilter' else (n_freqs,)
    n_timepts = method_to_n_timepts[method]

    # Basic test of shape, dtype, value of output.
    # Test values averaged over all timepts, freqs for 1st trial for simplicity
    spec, freqs, timepts = itpc(data, smp_rate, axis=0, method=method, itpc_method=itpc_method,
                                trial_axis=-1)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert spec.shape == (n_freqs, n_timepts)
    assert np.isclose(spec[:,:].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    spec, freqs, timepts = itpc(np.tile(data[:,:,np.newaxis],(1,1,2)), smp_rate, axis=0,
                                method=method, itpc_method=itpc_method, trial_axis=1)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert spec.shape == (n_freqs, n_timepts, 2)
    assert np.isclose(spec[:,:,0].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with transposed data dimensionality
    spec, freqs, timepts = itpc(data.T, smp_rate, axis=-1, method=method, itpc_method=itpc_method,
                                trial_axis=0)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert spec.shape == (n_freqs, n_timepts)
    assert np.isclose(spec[:,:].mean(), result, rtol=1e-4, atol=1e-4)

    # Test for expected output with time-reversed data
    # Skip test for bandfilter method -- different init conds do change results slightly at start
    # Skip test for multitaper method -- not time-reversal invariant due to windowing
    if method not in ['bandfilter','multitaper']:
        spec, freqs, timepts = itpc(np.flip(data,axis=0), smp_rate, axis=0, method=method,
                                    itpc_method=itpc_method, trial_axis=-1)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert spec.shape == (n_freqs, n_timepts)
        assert np.isclose(spec[:,:].mean(), result, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        spec, freqs, timepts = itpc(data, smp_rate, axis=0, method=method, itpc_method=itpc_method,
                                    trial_axis=-1, foo=None)


# =============================================================================
# Unit tests for spectral/LFP preprocessing functions
# =============================================================================
def test_cut_trials(oscillation):
    """ Unit tests for cut_trials function """
    n_timepts, n_trials = oscillation.shape

    # Unwrap trial-cut data into one long vector to simulate uncut data
    uncut_data  = oscillation.flatten(order='F')
    data        = oscillation
    data_orig   = uncut_data.copy()

    # Check if unwrapped->recut data is same as original uncut data
    trial_lims  = np.asarray([0,0.999])[np.newaxis,:] + np.arange(n_trials)[:,np.newaxis]
    cut_data    = cut_trials(uncut_data, trial_lims, smp_rate=1000, axis=0)
    assert np.array_equal(uncut_data,data_orig)     # Ensure input data isn't altered by function
    assert cut_data.shape == data.shape
    assert (cut_data == data).all()

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        cut_data = cut_trials(uncut_data, trial_lims, smp_rate=1000, axis=0, foo=None)


def test_realign_data(oscillation):
    """ Unit tests for realign_data function """
    data = oscillation
    data_orig = data.copy()
    n_timepts, n_trials = data.shape
    timepts = np.arange(0,1,1e-3)

    # Realign to 2 distinct times, then concatenate together and test if same
    realigned1 = realign_data(data, 0.5*np.ones((n_trials,)), time_range=(-0.5,-0.001),
                                timepts=timepts, time_axis=0, trial_axis=-1)
    realigned2 = realign_data(data, 0.5*np.ones((n_trials,)), time_range=(0,0.499),
                                timepts=timepts, time_axis=0, trial_axis=-1)
    realigned = np.concatenate((realigned1,realigned2), axis=0)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert realigned.shape == data.shape
    assert (realigned == data).all()

    # Test for consistent output with transposed data dimensionality
    realigned1 = realign_data(data.T, 0.5*np.ones((n_trials,)), time_range=(-0.5,-0.001),
                                timepts=timepts, time_axis=-1, trial_axis=0)
    realigned2 = realign_data(data.T, 0.5*np.ones((n_trials,)), time_range=(0,0.499),
                                timepts=timepts, time_axis=-1, trial_axis=0)
    realigned = np.concatenate((realigned1,realigned2), axis=-1)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert realigned.shape == data.T.shape
    assert (realigned == data.T).all()

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        realigned = realign_data(data, 0.5*np.ones((n_trials,)), time_range=(-0.5,-0.001),
                                 timepts=timepts, time_axis=0, trial_axis=-1, foo=None)


@pytest.mark.parametrize('axis', [0, None])
def test_remove_dc(oscillation, axis):
    """ Unit tests for remove_dc() function """
    data = oscillation
    data_orig = data.copy()
    
    # Test that DC-removed data does have mean=0
    data_no_dc = remove_dc(data, axis=axis)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(data_no_dc.shape, data.shape)
    if axis == None:    assert np.isclose(data_no_dc.sum(), 0)
    else:               assert np.allclose(data_no_dc.sum(axis=axis), 0)

    # Test for consistency with transposed data
    axis_T = 1 if axis == 0 else None
    data_no_dc = remove_dc(data.T, axis=axis_T)
    assert np.array_equal(data_no_dc.shape, data.T.shape)
    if axis == None:    assert np.isclose(data_no_dc.sum(), 0)
    else:               assert np.allclose(data_no_dc.sum(axis=axis_T), 0)
    
    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        data_no_dc = remove_dc(data, axis=axis, foo=None)
        
  
@pytest.mark.parametrize('method, result, result2',
                         [('mean',      13.3, 0.31),
                          ('groupmean', 0.49, 13.13),
                          ('regress',   0.49, 13.13)])
def test_remove_evoked(oscillation, method, result, result2):
    """ Unit tests for remove_evoked() function """
    data = oscillation
    data[:,2:] = np.roll(data[:,2:], shift=15, axis=0)  # Shift last 2 simulated trials by ~180 deg
    data_orig = data.copy()

    if method == 'mean':        design = None
    elif method == 'groupmean': design = [0,0,1,1]
    elif method == 'regress':   design = np.vstack(([0,0,0,0], [-1,-1,1,1])).T
                
    # Basic test of function
    data_no_evoked, evoked = remove_evoked(data, axis=1, method=method, design=design, return_evoked=True)
    print(np.round(power(data_no_evoked).mean(),2), np.round(power(evoked).mean(),2))
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(data_no_evoked.shape, data.shape)
    assert np.array_equal(evoked.shape, data.shape)
    assert np.isclose(power(data_no_evoked).mean(), result, rtol=1e-2, atol=1e-2)
    assert np.isclose(power(evoked).mean(), result2, rtol=1e-2, atol=1e-2)

    # Test for consistency with not returning evoked potential
    data_no_evoked = remove_evoked(data, axis=1, method=method, design=design, return_evoked=False)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(data_no_evoked.shape, data.shape)
    assert np.isclose(power(data_no_evoked).mean(), result, rtol=1e-2, atol=1e-2)

    # Test for consistency with transposed data
    data_no_evoked = remove_evoked(data.T, axis=0, method=method, design=design)
    assert np.array_equal(data_no_evoked.shape, data.T.shape)
    assert np.isclose(power(data_no_evoked).mean(), result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        data_no_evoked = remove_evoked(data, axis=0, method=method, foo=None)


# =============================================================================
# Unit tests for spectral/LFP postprocessing functions
# =============================================================================
def test_one_over_f_norm():
    """ Unit tests for one_over_f_norm() function """    
    # Generate idealized 1/f spectrogam ~ (n_freqs,n_trials) = (100,4)
    f = iarange(1,100)
    data = np.tile((1.0/f)[:,np.newaxis], (1,4))
    data_orig = data.copy()

    # Basic test of function
    data_corrected = one_over_f_norm(data, axis=0, freqs=f, exponent=1.0)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(data_corrected.shape, data.shape)
    assert np.allclose(data_corrected.std(axis=0), 0)
    
    # Test for consistency for transposed data
    data_corrected = one_over_f_norm(data.T, axis=1, freqs=f, exponent=1.0)
    assert np.array_equal(data_corrected.shape, data.T.shape)
    assert np.allclose(data_corrected.std(axis=1), 0)

    # Test for expected output for higher exponent
    data = np.tile((1.0/f**2)[:,np.newaxis], (1,4))    
    data_corrected = one_over_f_norm(data, axis=0, freqs=f, exponent=2.0)
    assert np.allclose(data_corrected.std(axis=0), 0)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        data_corrected = one_over_f_norm(data, axis=0, freqs=f, foo=None)

             
@pytest.mark.parametrize('variable_type, pooler, result',
                         [('numpy',   'mean',   43.5538),
                          ('numpy',   'sum',    300.0944),
                          ('numpy',   'custom', 43.5538),
                          ('xarray',  'mean',   43.5538),
                          ('xarray',  'sum',    300.0944),
                          ('xarray',  'custom', 43.5538)])
def test_pool_freq_bands(oscillation, variable_type, pooler, result):
    """ Unit tests for pool_freq_bands function """
    data = oscillation
    n_timepts, n_trials = data.shape
    smp_rate = 1000
    timepts = np.arange(0,1,1e-3)

    # Set 'func' = callable custom pooling function
    if pooler == 'custom': pooler = lambda x: np.mean(x,axis=0)

    variable_type_ = xr.DataArray if variable_type == 'xarray' else np.ndarray

    spec, freqs, timepts = spectrogram(data, smp_rate, axis=0, method='wavelet',
                                       data_type='lfp', spec_type='power')
    # Convert Numpy ndarray -> xarray DataArray
    if variable_type == 'xarray':
        spec = xr.DataArray(spec,
                            dims=['frequency','time','trial'],
                            coords={'frequency':freqs, 'time':timepts, 'trial':np.arange(n_trials)})
    spec_orig = spec.copy()

    # Frequency bands to pool data within
    bands   = OrderedDict({'theta':[3,8], 'beta':[10,32], 'gamma':[40,100]})
    n_bands = len(bands)

    extra_args = dict(axis=0, freqs=freqs) if variable_type == 'numpy' else {}

    eval_func = (lambda x: x.mean().values) if variable_type == 'xarray' else (lambda x: x.mean())

    # Basic test of shape, type, value of output.
    # Test values averaged over all timepts, freqs for 1st trial for simplicity
    band_spec = pool_freq_bands(spec, bands, func=pooler, **extra_args)
    # print(band_spec[:,:,0].mean(), np.mean(band_spec[:,:,0]))
    # print(eval_func(band_spec[:,:,0]))
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    assert isinstance(band_spec,variable_type_)
    assert band_spec.shape == (n_bands, n_timepts, n_trials)
    assert np.isclose(eval_func(band_spec[:,:,0]), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    spec2 = np.tile(spec[:,:,:,np.newaxis],(1,1,2)) if variable_type == 'numpy' else \
            xr.concat((spec,spec), dim='channel').transpose('frequency','time','trial','channel')
    band_spec = pool_freq_bands(spec2, bands, func=pooler, **extra_args)
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    assert band_spec.shape == (n_bands, n_timepts, n_trials, 2)
    assert np.isclose(eval_func(band_spec[:,:,0,0]), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with transposed data dimensionality
    if variable_type == 'numpy': extra_args['axis'] = 2
    band_spec = pool_freq_bands(spec.transpose(), bands, func=pooler, **extra_args)
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    assert band_spec.shape == (n_trials, n_timepts, n_bands)
    assert np.isclose(eval_func(band_spec[0,:,:]), result, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        band_spec = pool_freq_bands(spec, bands, func=pooler, foo=None, **extra_args)


@pytest.mark.parametrize('variable_type, pooler, result',
                         [('numpy',   'mean',   44.9441),
                          ('numpy',   'sum',    22494.0313),
                          ('numpy',   'custom', 44.9441),
                          ('xarray',  'mean',   44.9441),
                          ('xarray',  'sum',    22494.0313),
                          ('xarray',  'custom', 44.9441)])
def test_pool_time_epochs(oscillation, variable_type, pooler, result):
    """ Unit tests for pool_time_epochs function """
    data = oscillation
    n_timepts, n_trials = data.shape
    smp_rate = 1000
    timepts = np.arange(0,1,1e-3)

    # Set 'func' = callable custom pooling function
    if pooler == 'custom': pooler = lambda x: np.mean(x,axis=0)

    variable_type_ = xr.DataArray if variable_type == 'xarray' else np.ndarray

    spec, freqs, timepts = spectrogram(data, smp_rate, axis=0, method='wavelet',
                                       data_type='lfp', spec_type='power')
    # Convert Numpy ndarray -> xarray DataArray
    if variable_type == 'xarray':
        spec = xr.DataArray(spec,
                            dims=['frequency','time','trial'],
                            coords={'frequency':freqs, 'time':timepts, 'trial':np.arange(n_trials)})
    spec_orig = spec.copy()
    n_freqs = len(freqs)

    # Make up some time epochs to pool data within
    epochs   = OrderedDict({'sample':[0,0.5], 'delay':[0.5,1]})
    n_epochs = len(epochs)

    extra_args = dict(axis=1, timepts=timepts) if variable_type == 'numpy' else {}

    eval_func = (lambda x: x.mean().values) if variable_type == 'xarray' else (lambda x: x.mean())

    # Basic test of shape, type, value of output.
    # Test values averaged over all timepts, freqs for 1st trial for simplicity
    epoch_spec = pool_time_epochs(spec, epochs, func=pooler, **extra_args)
    print(epoch_spec[:,:,0].mean(), np.mean(epoch_spec[:,:,0]))
    print(eval_func(epoch_spec[:,:,0]))
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    assert isinstance(epoch_spec,variable_type_)
    assert epoch_spec.shape == (n_freqs, n_epochs, n_trials)
    assert np.isclose(eval_func(epoch_spec[:,:,0]), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    spec2 = np.tile(spec[:,:,:,np.newaxis],(1,1,2)) if variable_type == 'numpy' else \
            xr.concat((spec,spec), dim='channel').transpose('frequency','time','trial','channel')
    epoch_spec = pool_time_epochs(spec2, epochs, func=pooler, **extra_args)
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    assert epoch_spec.shape == (n_freqs, n_epochs, n_trials, 2)
    assert np.isclose(eval_func(epoch_spec[:,:,0,0]), result, rtol=1e-4, atol=1e-4)

    # Test for consistent output with transposed data dimensionality
    if variable_type == 'numpy': extra_args['axis'] = 2
    spec2 = spec.transpose((0,2,1)) if variable_type == 'numpy' else \
            spec.transpose('frequency','trial','time')
    epoch_spec = pool_time_epochs(spec2, epochs, func=pooler, **extra_args)
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    assert epoch_spec.shape == (n_freqs, n_trials, n_epochs)
    assert np.isclose(eval_func(epoch_spec[:,0,:]), result, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        epoch_spec = pool_time_epochs(spec, epochs, func=pooler, foo=None, **extra_args)


# =============================================================================
# Unit tests for spectral plotting functions
# =============================================================================
@pytest.mark.parametrize('method', [('multitaper'), ('wavelet'), ('bandfilter')])
def test_plot_spectrum(oscillation, method):
    """ Unit tests for plot_spectrum function with different frequency sampling """
    data = oscillation
    smp_rate = 1000

    spec, freqs = spectrum(data, smp_rate, axis=0, method=method, spec_type='power')
    mean = spec.mean(axis=-1)
    mean_orig = mean.copy()
    freqs_orig = freqs.copy()
    if method == 'wavelet':         freqs_result = np.log2(freqs)
    elif method == 'bandfilter':    freqs_result = np.arange(len(freqs))
    else:                           freqs_result = freqs

    # Basic test that plotted data == input data
    lines, _ = plot_spectrum(freqs, mean)
    assert np.array_equal(freqs, freqs_orig) # Ensure input data isn't altered by function
    assert np.array_equal(mean, mean_orig)
    assert np.allclose(lines[0].get_xdata(), freqs_result)
    assert np.allclose(lines[0].get_ydata(), mean)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        lines, _ = plot_spectrum(freqs, mean, foo=None)


@pytest.mark.parametrize('method', [('multitaper'), ('wavelet'), ('bandfilter')])
def test_plot_spectrogram(oscillation, method):
    """ Unit tests for plot_spectrogram function with different frequency sampling """
    data = oscillation
    smp_rate = 1000

    spec, freqs, timepts = spectrogram(data, smp_rate, axis=0, method=method, spec_type='power')
    spec = spec.mean(axis=-1)
    freqs_orig = freqs.copy()
    spec_orig = spec.copy()

    # Basic test that plotted data == input data
    img, _ = plot_spectrogram(timepts, freqs, spec)
    assert np.array_equal(freqs, freqs_orig) # Ensure input data isn't altered by function
    assert np.array_equal(spec, spec_orig)
    assert np.allclose(img.get_array().data, spec)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        img, _ = plot_spectrogram(timepts, freqs, spec, foo=None)


# =============================================================================
# Unit tests for other spectral/LFP utility functions
# =============================================================================
def test_get_freq_sampling():
    """ Unit tests for get_freq_sampling() function """
    smp_rate = 1000
    nfft = 1000
    
    # Basic test of function
    f, fbool = get_freq_sampling(smp_rate, nfft)
    print(f.shape, f[:10], f[-1])
    assert np.array_equal(f, iarange(0,500))

    # Test with different freq_range, nfft, smp_rate
    f, fbool = get_freq_sampling(smp_rate, nfft, freq_range=(100,200))
    assert np.array_equal(f, iarange(100,200))

    f, fbool = get_freq_sampling(smp_rate, 500)
    assert np.array_equal(f, iarange(0,500,2))

    f, fbool = get_freq_sampling(smp_rate, nfft, two_sided=True)
    assert np.array_equal(f, np.hstack((iarange(0,499), iarange(-500,-1))))

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        f, fbool = get_freq_sampling(smp_rate, nfft, foo=None)


def test_one_sided_to_two_sided(oscillation):
    """ Unit tests for one_sided_to_two_sided() function """
    data = oscillation
    smp_rate = 1000
    n_trials = 4

    # Basic test of shape, dtype, value of output.
    # Test values averaged over all timepts, freqs for 1st trial for simplicity
    spec, freqs, timepts = spectrogram(data, smp_rate, axis=0, method='multitaper',
                                       data_type='lfp', spec_type='complex')    
    spec_orig = spec.copy()
    
    # TODO Set up some basic tests of function
    spec2 = one_sided_to_two_sided(spec, freqs, smp_rate, axis=0)
    assert np.array_equal(spec,spec_orig)     # Ensure input data isn't altered by function
    
    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        spec2 = one_sided_to_two_sided(spec, freqs, smp_rate, axis=0, foo=None)
    
    
def test_imports():
    """ Test different import methods for spectra module """
    # Import entire package
    import spynal
    spynal.spectra.wavelet.wavelet_spectrogram    
    spynal.spectra.wavelet_spectrogram
    # Import module
    import spynal.spectra as spec
    spec.wavelet.wavelet_spectrogram
    spec.wavelet_spectrogram
    # Import specific function from module
    from spynal.spectra import wavelet_spectrogram
    wavelet_spectrogram
    # Import specific function from module
    from spynal.spectra.wavelet import wavelet_spectrogram
    wavelet_spectrogram
        


# -*- coding: utf-8 -*-
"""
Random sample generators for randomization, permutation (shuffle), and bootstrap statistics

Function reference
------------------
- permutations :                Generate random permutations (resampling w/o replacement)
- bootstraps :                  Generate random bootstrap samples (resampling w/ replacement)
- signs :                       Generate random binary variables (eg for sign tests)
- jackknifes :                  Generate jackknife samples (exclude each observation in turn)
- subsets :                     Generate random length-k subsets (resampling w/o replacement)

Function reference
------------------
"""
import numpy as np

from spynal.utils import set_random_seed


# =============================================================================
# Random sample generators
# =============================================================================
def permutations(n, n_resamples=9999, seed=None):
    """
    Yield generator with a set of `n_resamples` random permutations of integers
    0:n-1, as would be needed for permutation/randomization tests

    Parameters
    ----------
    n : int
        Number of items to randomly resample from.
        Will usually correspond to number of observations/trials

    n_resamples : int, default: 9999 (appropriate number for test w/ 10,000 samples)
        Number of independent resamples to generate.

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    Yields
    ------
    resamples : generator, shape=(n_resamples,) of [ndarray, shape=(n,), dtype=int]
        Generator to iterate over for permutation test.
        Each iteration contains a random permutation of integers 0:n-1.
    """
    if seed is not None: set_random_seed(seed)

    for _ in range(n_resamples):
        yield np.random.permutation(n)


def bootstraps(n, n_resamples=9999, seed=None):
    """
    Yield generator with a set of `n_resamples` random resamplings with
    replacement of integers 0:n-1, as would be needed for bootstrap tests or
    confidence intervals

    Parameters
    ----------
    n : int
        Number of items to randomly resample from.
        Will usually correspond to number of observations/trials

    n_resamples : int, default: 9999 (appropriate number for test w/ 10,000 samples)
        Number of independent resamples to generate.

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    Yields
    ------
    resamples : generator, shape=(n_resamples,) of [ndarray, shape=(n,), dtype=int]
        Generator to iterate over for boostrap test or confidence interval computation.
        Each iteration contains a random resampling with replacement from integers 0:n-1.
    """
    if seed is not None: set_random_seed(seed)

    for _ in range(n_resamples):
        yield np.random.randint(n, size=(n,))


def signs(n, n_resamples=9999, seed=None):
    """
    Yield generator with a set of `n_resamples` random Bernoulli(p=0.5)
    variables (ie binary 0/1 w/ probability of 0.5), each of length <n>,
    as would be needed to set the signs of stats in a sign test.

    Parameters
    ----------
    n : int
        Number of items to randomly resample from.
        Will usually correspond to number of observations/trials

    n_resamples : int, default: 9999 (appropriate number for test w/ 10,000 samples)
        Number of independent resamples to generate.

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    Yields
    ------
    resamples : generator, shape=(n_resamples,) of [ndarray, shape=(n,), dtype=bool]
        Generator to iterate over for random sign test.
        Each iteration contains a random resampling of n Bernoulli random variables.
    """
    if seed is not None: set_random_seed(seed)

    for _ in range(n_resamples):
        yield np.random.binomial(1,0.5, size=(n,)).astype(bool)


def jackknifes(n, n_resamples=None, seed=None):
    """
    Yield generator with a set of n_resamples = n boolean variables,
    each of length n, and each of which excludes one observation/trial in turn,
    as would be needed for a jackknife or leave-one-out test.

    Parameters
    ----------
    n : int
        Number of items to randomly resample from.
        Will usually correspond to number of observations/trials

    n_resamples : int
        Automatically set=n here. Only included for consistent interface.

    seed : None
        Not used. Only included for consistent interface with other functions.

    Yields
    ------
    resamples : generator, shape=(n,) of [ndarray, shape=(n,), dtype=bool]
        Generator to iterate over for jackknife test.
        Each iteration is all 1's except for a single 0, the observation (trial) excluded
        in that iteration. For the ith resample, the ith trial is excluded.
    """
    assert (n_resamples is None) or (n_resamples == n), \
        ValueError("For jackknife/leave-one-out, n_resamples MUST = n")

    trials = np.arange(n)
    for trial in range(n):
        yield trials != trial


def subsets(n, k, n_resamples=9999, seed=None):
    """
    Yield generator with a set of `n_resamples` random length-k subsets of integers 0:n-1.
    
    Random sampling version of "n-choose-k" function.
    
    Parameters
    ----------
    n : int
        Number of items to randomly resample from.
        Will usually correspond to number of observations/trials.

    k : int
        Length of subset to select. For example, k=2 implies random sampling from "n-choose-2".

    n_resamples : int, default: 9999 (appropriate number for test w/ 10,000 samples)
        Number of independent resamples to generate.

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    Yields
    ------
    resamples : generator, shape=(n_resamples,) of [ndarray, shape=(k,), dtype=int]
        Generator to iterate over for a random-subset test.
        Each iteration contains a random subset of integers 0:n-1.
    """
    if seed is not None: set_random_seed(seed)

    for _ in range(n_resamples):
        yield np.random.permutation(n)[:k]



import setuptools
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), 'r') as f:
    long_description = f.read()

requirements = [
        'numpy',
        'scipy',
        'pandas',
        'matplotlib',
        'h5py',
        'hdf5storage',
        'patsy',
        'scikit-learn',
        'pyfftw>=0.12.0'
        ]

setuptools.setup(
      name='spynal',
      version='0.1.3',
      description='Simple Python Neural Analysis Library',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/sbrincat/spynal.git',
      author= ['Scott Brincat', 'John Tauber'],
      author_email= 'sbrincat@mit.edu',
      license='LICENSE',
      packages=setuptools.find_packages(),
      install_requires=requirements,
      classifiers = [
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.7',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS',
      ],
      zip_safe=False)


""" Unit tests for sync.py module """
from math import pi
import pytest
import numpy as np

from scipy.stats import bernoulli

from spynal.tests.data_fixtures import MISSING_ARG_ERRS
from spynal.utils import index_axis
from spynal.sync.sync import simulate_multichannel_oscillation, synchrony, spike_field_coupling
from spynal.spectra.spectra import spectrogram

# =============================================================================
# Fixtures for generating simulated data
# =============================================================================
@pytest.fixture(scope='session')
def oscillation_pair():
    """
    Fixture simulates set of instances of pairs of weakly synchronized oscillatory data
    for unit tests of field-field synchrony computation functions

    RETURNS
    data    (1000,40,2) ndarray. Simulated oscillatory data.
            (eg simulating 1000 timepoints x 40 trials x 2 channels)
    """
    # Note: seed=1 makes data reproducibly match output of Matlab
    frequency = 32
    return simulate_multichannel_oscillation(2, frequency, amplitude=5.0, phase=[pi/4,0],
                                             phase_sd=[0,pi/4], noise=1.0, n_trials=40,
                                             time_range=1.0, smp_rate=1000, seed=1)


@pytest.fixture(scope='session')
def spike_field_pair(oscillation_pair):
    """
    Fixture simulates set of instances of weakly synchronized oscillatory spike-field data pairs
    for unit tests of spike-field synchrony computation functions

    RETURNS
    spkdata (1000,40) ndarray of bool. Simulated oscillatory spiking data,
            expressed as binary (0/1) spike trains.
            (eg simulating 1000 timepoints x 4 trials or channels)

    lfpdata (1000,40) ndarray of float. Simulated oscillatory LFP data
            (eg simulating 1000 timepoints x 4 trials or channels)
    """
    # todo code up something actually proper (rate-modulated Poisson process?)
    spkdata,lfpdata = oscillation_pair[:,:,0], oscillation_pair[:,:,1]

    # Convert continuous oscillation to probability (range 0-1)
    spkdata = (spkdata - spkdata.min()) / spkdata.ptp()
    spkdata = spkdata**2  # Sparsen high rates some

    # Use probabilities to generate Bernoulli random variable at each time point
    spkdata =  bernoulli.ppf(0.5, spkdata).astype(bool)

    return spkdata, lfpdata


# =============================================================================
# Unit tests
# =============================================================================
@pytest.mark.parametrize('method, spec_method, single_trial, result',
                         [('coherence', 'wavelet',      None,       (0.2475,0.4795)),
                          ('coherence', 'multitaper',   None,       (0.1046,0.3043)),
                          ('coherence', 'bandfilter',   None,       (0.3490,0.6702)),
                          ('PLV',       'wavelet',      None,       (0.2541,0.3713)),
                          ('PLV',       'multitaper',   None,       (0.0984,0.2228)),
                          ('PLV',       'bandfilter',   None,       (0.3321,0.4528)),
                          ('PPC',       'wavelet',      None,       (0.0912,0.3713)),
                          ('PPC',       'multitaper',   None,       (0.0100,0.2228)),
                          ('PPC',       'bandfilter',   None,       (0.1715,0.4528)),
                          ('coherence', 'wavelet',      'pseudo',   (0.1988,None)),
                          ('coherence', 'multitaper',   'pseudo',   (0.0623,None)),
                          ('coherence', 'bandfilter',   'pseudo',   (0.3080,None)),
                          ('PLV',       'wavelet',      'pseudo',   (0.2022,None)),
                          ('PLV',       'multitaper',   'pseudo',   (0.0588,None)),
                          ('PLV',       'bandfilter',   'pseudo',   (0.2858,None)),
                          ('PPC',       'wavelet',      'pseudo',   (0.0679,None)),
                          ('PPC',       'multitaper',   'pseudo',   (0.0015,None)),
                          ('PPC',       'bandfilter',   'pseudo',   (0.1502,None))])
def test_synchrony(oscillation_pair, method, spec_method, single_trial, result):
    """ Unit tests for synchrony() function """
    # Extract per-channel data and reshape -> (n_trials,n_timepts)
    data1, data2 = oscillation_pair[:,:,0].T, oscillation_pair[:,:,1].T
    data1_orig = data1.copy()
    data2_orig = data2.copy()

    smp_rate = 1000
    method_to_n_freqs   = {'wavelet': 26, 'multitaper':257,  'bandfilter': 3}
    method_to_n_timepts = {'wavelet': 1000, 'multitaper':2,  'bandfilter': 1000}
    n_freqs     = method_to_n_freqs[spec_method]
    n_timepts   = method_to_n_timepts[spec_method]
    freqs_shape = (n_freqs,2) if spec_method == 'bandfilter' else (n_freqs,)
    n_trials    = data1.shape[0]
    sync_shape  = (n_trials, n_freqs, n_timepts) if single_trial else (1, n_freqs, n_timepts)
    if spec_method == 'multitaper': sync_shape = (*sync_shape[:2], 1, sync_shape[-1])

    extra_args = dict(axis=0, method=method, spec_method=spec_method,
                      single_trial=single_trial,
                      return_phase=True if single_trial is None else False,
                      smp_rate=smp_rate, time_axis=-1)

    # Basic test of shape, dtype, value of output.
    # Test values averaged over all timepts, freqs for simplicity
    if single_trial:    sync, freqs, timepts = synchrony(data1, data2, **extra_args)
    else:               sync, freqs, timepts, dphi = synchrony(data1, data2, **extra_args)
    print(np.round(sync.mean(),4))
    if not single_trial: print(np.round(dphi.mean(),4))
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    assert sync.shape == sync_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.isclose(sync.mean(), result[0], rtol=1e-4, atol=1e-4)
    if not single_trial:
        assert np.issubdtype(dphi.dtype,float)
        assert np.isclose(dphi.mean(), result[1], rtol=1e-4, atol=1e-4)

    # Test for consistent output with return_phase=False
    # Note: Returning phase not allowed for single-trial estimate, so skip in that case
    if not single_trial:
        extra_args['return_phase'] = False
        sync2, freqs2, timepts2 = synchrony(data1, data2, **extra_args)
        assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
        assert np.array_equal(data2,data2_orig)
        assert freqs2.shape == freqs_shape
        assert timepts2.shape == (n_timepts,)
        assert np.allclose(sync2, sync, rtol=1e-4, atol=1e-4)
        extra_args['return_phase'] = True

    # Test for consistent output with reversed time axis (dphi should change sign, otherwise same)
    # Skip test for multitaper/bandfilter phase, bandfilter sync -- not time-reversal invariant
    if single_trial:
        sync, freqs, timepts = synchrony(np.flip(data1,axis=-1), np.flip(data2,axis=-1),
                                         **extra_args)
    else:
        sync, freqs, timepts, dphi = synchrony(np.flip(data1,axis=-1), np.flip(data2,axis=-1),
                                               **extra_args)
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    # HACK Bandfilter not time-reversal invariant due to initial conditions
    if spec_method != 'bandfilter':
        assert np.isclose(sync.mean(), result[0], rtol=1e-4, atol=1e-4)
    # HACK Only synchrony magnitude -- not phase -- is time-reversal invariant for multitaper
    if not single_trial and (spec_method == 'wavelet'):
        assert np.isclose(dphi.mean(), -result[1], rtol=1e-4, atol=1e-4)

    # Test for consistent output with channels swapped (dphi should change sign, otherwise same)
    # Skip test for multitaper/bandfilter phase -- not time-reversal invariant
    if single_trial:    sync, freqs, timepts = synchrony(data2, data1, **extra_args)
    else:               sync, freqs, timepts, dphi = synchrony(data2, data1, **extra_args)
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    assert np.isclose(sync.mean(), result[0], rtol=1e-4, atol=1e-4)
    if not single_trial and (spec_method == 'wavelet'):
        assert np.isclose(dphi.mean(), -result[1], rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    extra_args['time_axis'] = 1
    if single_trial:
        sync, freqs, timepts = synchrony(np.stack((data1,data1),axis=2),
                                         np.stack((data2,data2),axis=2), **extra_args)
    else:
        sync, freqs, timepts, dphi = synchrony(np.stack((data1,data1),axis=2),
                                               np.stack((data2,data2),axis=2), **extra_args)
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert sync.shape == (*sync_shape, 2)
    assert np.issubdtype(sync.dtype,float)
    # HACK Loosen tolerance bc this doesn't quite match up for bandfilter (TODO why???)
    if single_trial:
        assert np.isclose(sync[...,0].mean(), result[0], rtol=1e-3, atol=1e-3)
    else:
        assert np.isclose(sync[...,0].mean(), result[0], rtol=1e-3, atol=1e-3)
        assert dphi.shape == (*sync_shape, 2)
        assert np.issubdtype(dphi.dtype,float)
        assert np.isclose(dphi[...,0].mean(), result[1], rtol=1e-3, atol=1e-3)
    extra_args['time_axis'] = -1

    # Test for consistent output with transposed data dimensionality -> (time,trials)
    transposed_shape = (*sync_shape[1:],sync_shape[0])
    extra_args.update(axis=-1, time_axis=0)
    if single_trial:    sync, freqs, timepts = synchrony(data1.T, data2.T, **extra_args)
    else:               sync, freqs, timepts, dphi = synchrony(data1.T, data2.T, **extra_args)
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert sync.shape == transposed_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.isclose(sync.mean(), result[0], rtol=1e-4, atol=1e-4)
    if not single_trial:
        assert np.issubdtype(dphi.dtype,float)
        assert dphi.shape == transposed_shape
        assert np.isclose(dphi.mean(), result[1], rtol=1e-4, atol=1e-4)
    extra_args.update(axis=0, time_axis=-1)

    # Test for consistent output with keepdims=False (reduced trial axis)
    reduced_shape = tuple([j for j in sync_shape if j != 1])
    if single_trial:
        sync, freqs, timepts = synchrony(data1, data2, keepdims=False, **extra_args)
    else:
        sync, freqs, timepts, dphi = synchrony(data1, data2, keepdims=False, **extra_args)
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    assert sync.shape == reduced_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.isclose(sync.mean(), result[0], rtol=1e-4, atol=1e-4)
    if not single_trial:
        assert np.issubdtype(dphi.dtype,float)
        assert dphi.shape == reduced_shape
        assert np.isclose(dphi.mean(), result[1], rtol=1e-4, atol=1e-4)

    # Test for consistent output with spectral data input
    spec_args = dict(axis=1, method=spec_method, spec_type='complex')
    if spec_method == 'multitaper': spec_args.update(keep_tapers=True)
    spec1, freqs, timepts = spectrogram(data1, smp_rate, **spec_args)
    spec2, freqs, timepts = spectrogram(data2, smp_rate, **spec_args)
    extra_args['taper_axis'] = 2
    if single_trial:    sync, _, _ = synchrony(spec1, spec2, **extra_args)
    else:               sync, _, _, dphi = synchrony(spec1, spec2, **extra_args)
    assert np.array_equal(data1,data1_orig)     # Ensure input data not altered by func
    assert np.array_equal(data2,data2_orig)
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert sync.shape == sync_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.isclose(sync.mean(), result[0], rtol=1e-4, atol=1e-4)
    if not single_trial:
        assert np.issubdtype(dphi.dtype,float)
        assert dphi.shape == sync_shape
        assert np.isclose(dphi.mean(), result[1], rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        extra_args['foo'] = None
        if single_trial:    sync, freqs, timepts = synchrony(data1, data2, **extra_args)
        else:               sync, freqs, timepts, dphi = synchrony(data1, data2, **extra_args)


@pytest.mark.parametrize('method, spec_method, result',
                         [('coherence', 'wavelet',      (0.2399,0.2941)),
                          ('coherence', 'multitaper',   (0.1218,0.1556)),
                          ('coherence', 'bandfilter',   (0.3722,0.2734)),
                          ('PLV',       'wavelet',      (0.1600,-0.2388,1994)),
                          ('PLV',       'multitaper',   (0.0576,0.9267,6031)),
                          ('PLV',       'bandfilter',   (0.3040,-0.2257,1994)),
                          ('PPC',       'wavelet',      (0.0764,-0.2388,1994)),
                          ('PPC',       'multitaper',   (0.0085,0.9267,6031)),
                          ('PPC',       'bandfilter',   (0.1417,-0.2257,1994))])
def test_spike_field_coupling(spike_field_pair, method, spec_method, result):
    """ Unit tests for spike_field_coupling() function """
    # Extract per-channel data and reshape -> (n_trials,n_timepts)
    spkdata, lfpdata = spike_field_pair[0].T, spike_field_pair[1].T
    spkdata_orig = spkdata.copy()
    lfpdata_orig = lfpdata.copy()

    smp_rate = 1000
    method_to_n_freqs   = {'wavelet': 26, 'multitaper':129,  'bandfilter': 3}
    method_to_n_timepts = {'wavelet': 1000, 'multitaper':5,  'bandfilter': 1000} \
                          if method == 'coherence' else \
                          {'wavelet': 4, 'multitaper':4,  'bandfilter': 4}
    n_freqs     = method_to_n_freqs[spec_method]
    n_timepts   = method_to_n_timepts[spec_method]
    freqs_shape = (n_freqs,2) if spec_method == 'bandfilter' else (n_freqs,)
    timepts     = np.arange(lfpdata.shape[-1]) / smp_rate
    sync_shape  = (1, n_freqs, n_timepts)
    # DEL if spec_method == 'multitaper': sync_shape = (*sync_shape[:2], sync_shape[-1])    
    if spec_method == 'multitaper': sync_shape = (*sync_shape[:2], 1, sync_shape[-1])

    extra_args  = {'timepts':timepts, 'width':0.2} if method != 'coherence' else {}
    if spec_method == 'multitaper':
        extra_args.update(time_width=0.2, spacing=0.2, freq_width=10)

    # Basic test of shape, dtype, value of output.
    # Test values averaged over all timepts, freqs for simplicity
    sync, freqs, timepts, n, phi = spike_field_coupling(spkdata, lfpdata, axis=0, time_axis=-1,
                                                        method=method, spec_method=spec_method,
                                                        smp_rate=smp_rate, return_phase=True,
                                                        **extra_args)
    print("TEST1", sync.mean(), phi.mean(), result)
    assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
    assert np.array_equal(lfpdata,lfpdata_orig)
    assert isinstance(freqs, np.ndarray)
    assert isinstance(timepts, np.ndarray)
    assert isinstance(sync, np.ndarray)
    assert isinstance(phi, np.ndarray)
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert sync.shape == sync_shape
    assert phi.shape == sync_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.issubdtype(phi.dtype,float)
    assert np.isclose(np.nanmean(sync), result[0], rtol=1e-4, atol=1e-4)
    # assert np.isclose(np.nanmean(phi), result[1], rtol=1e-4, atol=1e-4)
    if method != 'coherence':
        assert isinstance(n, np.ndarray)
        assert n.shape == (n_timepts,)
        assert np.round(n.mean()) == result[2]


    # Test for consistent output with return_phase=False
    sync2, freqs2, timepts2, n = spike_field_coupling(spkdata, lfpdata, axis=0, time_axis=-1,
                                                      method=method, spec_method=spec_method,
                                                      smp_rate=smp_rate, return_phase=False,
                                                      **extra_args)
    assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
    assert np.array_equal(lfpdata,lfpdata_orig)
    assert freqs2.shape == freqs_shape
    assert timepts2.shape == (n_timepts,)
    assert np.isclose(np.nanmean(sync2), result[0], rtol=1e-4, atol=1e-4)
    if method != 'coherence': assert np.round(n.mean()) == result[2]

    # Test for consistent output with reversed time axis (phi should change sign, otherwise same)
    # Note: Due to windowing, results not time-reveral invariant for PLV/PPC, so skip those
    # Skip test for multitaper/bandfilter phase, bandfilter sync -- not time-reversal invariant
    sync, freqs, timepts, n, phi = spike_field_coupling(np.flip(spkdata,axis=-1),
                                                        np.flip(lfpdata,axis=-1),
                                                        axis=0, time_axis=-1, method=method,
                                                        spec_method=spec_method, smp_rate=smp_rate,
                                                        return_phase=True, **extra_args)
    assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
    assert np.array_equal(lfpdata,lfpdata_orig)
    assert sync.shape == sync_shape
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert phi.shape == sync_shape
    # HACK Bandfilter not time-reversal invariant due to initial conditions
    if (spec_method != 'bandfilter') and (method not in ['PLV','PPC']):
        assert np.isclose(np.nanmean(sync), result[0], rtol=1e-4, atol=1e-4)
    # HACK Only synchrony magnitude -- not phase -- is time-reversal invariant for multitaper
    if (spec_method == 'wavelet') and (method not in ['PLV','PPC']):
        assert np.isclose(np.nanmean(phi), -result[1], rtol=1e-4, atol=1e-4)

    # Test for consistent output with different data array shape (3rd axis)
    if method == 'coherence':
        sync, freqs, timepts, n, phi = spike_field_coupling(np.stack((spkdata,spkdata),axis=2),
                                                            np.stack((lfpdata,lfpdata),axis=2),
                                                            axis=0, time_axis=1, method=method,
                                                            spec_method=spec_method,
                                                            smp_rate=smp_rate,
                                                            return_phase=True, **extra_args)
        assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
        assert np.array_equal(lfpdata,lfpdata_orig)
        assert freqs.shape == freqs_shape
        assert timepts.shape == (n_timepts,)
        assert sync.shape == (*sync_shape, 2)
        assert phi.shape == (*sync_shape, 2)
        print("SHAPE", sync[...,0].mean(), phi[...,0].mean(), sync.shape)
        assert np.issubdtype(sync.dtype,float)
        assert np.issubdtype(phi.dtype,float)
        assert np.isclose(sync[...,0].mean(), result[0], rtol=1e-4, atol=1e-4)
        # assert np.isclose(phi[...,0].mean(), result[1], rtol=1e-4, atol=1e-4)
        if method != 'coherence': assert np.round(n.mean()) == result[2]
        asdf
    # Test for consistent output with transposed data dimensionality -> (time,trials)
    transposed_shape = (*sync_shape[1:],sync_shape[0])
    sync, freqs, timepts, n, phi = spike_field_coupling(spkdata.T, lfpdata.T,
                                                        axis=-1, time_axis=0, method=method,
                                                        spec_method=spec_method, smp_rate=smp_rate,
                                                        return_phase=True, **extra_args)
    print(sync.shape, phi.shape)
    assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
    assert np.array_equal(lfpdata,lfpdata_orig)
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert sync.shape == transposed_shape
    assert phi.shape == transposed_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.issubdtype(phi.dtype,float)
    assert np.isclose(np.nanmean(sync), result[0], rtol=1e-4, atol=1e-4)
    assert np.isclose(np.nanmean(phi), result[1], rtol=1e-4, atol=1e-4)
    if method != 'coherence': assert np.round(n.mean()) == result[2]

    # Test for consistent output with spectral data input
    if spec_method == 'multitaper':
        spec_args = dict(time_width=0.2, freq_width=10, keep_tapers=True)
        if method != 'coherence': spec_args.update(spacing=1/smp_rate)
    else:
        spec_args = {}

    lfpspec, freqs, timepts = spectrogram(lfpdata, smp_rate, axis=1, method=spec_method,
                                          spec_type='complex', **spec_args)

    # Must convert spike data to spectral for coherence
    if method == 'coherence':
        spkspec, _, _ = spectrogram(spkdata, smp_rate, axis=1, method=spec_method,
                                    spec_type='complex', **spec_args)
    # Must insert singleton axis to match freq axis for other methods
    else:
        spkspec = spkdata[:,np.newaxis,np.newaxis,:] if spec_method == 'multitaper' \
                  else spkdata[:,np.newaxis,:]
    n_new_axes = 2 if spec_method == 'multitaper' else 1
    time_axis = 1 + n_new_axes

    extra_args_lcl = extra_args.copy()
    if (spec_method == 'multitaper') and (method != 'coherence'):
        timepts_raw = np.arange(0,lfpdata.shape[1])/smp_rate
        retained_times = (timepts_raw >= timepts[0]) & (timepts_raw <= timepts[-1])
        spkspec = index_axis(spkspec, time_axis, retained_times)
        extra_args_lcl.update(timepts=extra_args['timepts'][retained_times], taper_axis=time_axis-1)
    elif spec_method == 'multitaper':
        extra_args_lcl.update(timepts=timepts, taper_axis=time_axis-1, width=0.2)
    sync, _, _, n, phi = spike_field_coupling(spkspec, lfpspec, axis=0, time_axis=time_axis,
                                              method=method, spec_method=spec_method,
                                              return_phase=True, **extra_args_lcl)

    assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
    assert np.array_equal(lfpdata,lfpdata_orig)
    assert sync.shape == sync_shape
    assert phi.shape == sync_shape
    assert np.issubdtype(sync.dtype,float)
    assert np.issubdtype(phi.dtype,float)
    assert np.isclose(np.nanmean(sync), result[0], rtol=1e-4, atol=1e-4)
    assert np.isclose(np.nanmean(phi), result[1], rtol=1e-4, atol=1e-4)
    if method != 'coherence': assert np.round(n.mean()) == result[2]

    # Test for consistent output with keepdims=False (reduced trial axis)
    reduced_shape = tuple([j for j in sync_shape if j != 1])
    sync, freqs, timepts, n, phi = spike_field_coupling(spkdata, lfpdata, axis=0, time_axis=-1,
                                                        method=method, spec_method=spec_method,
                                                        smp_rate=smp_rate, return_phase=True,
                                                        keepdims=False, **extra_args)
    assert np.array_equal(spkdata,spkdata_orig)     # Ensure input data not altered by func
    assert np.array_equal(lfpdata,lfpdata_orig)
    assert sync.shape == reduced_shape
    assert freqs.shape == freqs_shape
    assert timepts.shape == (n_timepts,)
    assert phi.shape == reduced_shape
    assert np.isclose(np.nanmean(sync), result[0], rtol=1e-4, atol=1e-4)
    assert np.isclose(np.nanmean(phi), result[1], rtol=1e-4, atol=1e-4)
    if method != 'coherence':
        assert n.shape == (n_timepts,)
        assert np.round(n.mean()) == result[2]

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        sync, freqs, timepts, n, phi = spike_field_coupling(spkdata, lfpdata, axis=0, time_axis=-1,
                                                            method=method, spec_method=spec_method,
                                                            smp_rate=smp_rate, return_phase=True,
                                                            foo=None, **extra_args)


def test_imports():
    """ Test different import methods for sync module """
    # Import entire package
    import spynal
    spynal.sync.phasesync.ppc
    spynal.sync.ppc
    # Import module
    import spynal.sync as sync
    sync.phasesync.ppc
    sync.ppc
    # Import specific function from module
    from spynal.sync import ppc
    ppc
    # Import specific function from module
    from spynal.sync.phasesync import ppc
    ppc



""" Unit tests for info.py module """
import pytest
import numpy as np
import pandas as pd

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from patsy import dmatrix


from spynal.tests.data_fixtures import two_sample_data, one_way_data, two_way_data, \
                                       MISSING_ARG_ERRS
from spynal.info import neural_info, neural_info_2groups, neural_info_ngroups


# =============================================================================
# Unit tests
# =============================================================================
@pytest.mark.parametrize('method, params, result',
                         [('pev',           {'omega':True},         (34.43,35.91,71.03,60.72)),
                          ('pev',           {'omega':False},        (38.99,40.41,73.55,63.94)),
                          ('dprime',        {'signed':True},        (-1.52,-1.56,-3.16,-2.53)),
                          ('dprime',        {'signed':False},       (1.52,1.56,3.16,2.53)),
                          ('auroc',         {'signed':True},        (0.14,0.13,0.01,0.04)),
                          ('auroc',         {'signed':False},       (0.86,0.87,0.99,0.96)),
                          ('mutual_info',   {},                     (0.5,0.42,0.9,0.82)),
                          ('decode',        {'decoder':'LDA'},      0.90),
                          ('decode',        {'decoder':'logistic'}, 0.70),
                          ('decode',        {'decoder':'SVM'},      0.95),
                          ('decode',        {'decoder':'custom'},   0.90)])
def test_two_sample_info(two_sample_data, method, params, result):
    """ Unit tests for into functions for computing informatio for two-condition data """
    data, labels = two_sample_data
    data_orig = data.copy()

    n = int(10)
    n_chnls = int(4)
    info_shape = (1,n_chnls)
    # Set any extra parameters for info function
    extra_args = params
    if method == 'decode':
        extra_args['seed'] = 1
        # Test for consistency when passing custom decoder function
        if extra_args['decoder'] == 'custom':
            extra_args['decoder'] = LinearDiscriminantAnalysis(priors=(1/2)*np.ones((2,)))

    # Basic test of shape, value of output
    info = neural_info(data, labels, axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':
        print(round(info,2))
        assert isinstance(info,float)
    else:
        print(np.round(info.squeeze(),2))
        assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for expected output with keepdims = False
    info = neural_info(data, labels, axis=0, method=method, keepdims=False, **extra_args)
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == (n_chnls,)
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)    

    # Test for consistent output with 2-group form of neural computation function
    info = neural_info_2groups(data[labels==0,:], data[labels==1,:], axis=0, method=method,
                               **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with n-group form of neural computation function
    info = neural_info_ngroups(data[labels==0,:], data[labels==1,:], axis=0, method=method,
                               **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape (3rd axis)
    info = neural_info(data.reshape((n*2,int(n_chnls/2),int(n_chnls/2))), labels, axis=0,
                       method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert info.shape == (1,1,n_chnls/2)
    else:                   assert info.shape == (1,n_chnls/2,n_chnls/2)
    # Skip decoding bc different results here due to different # of channels (decoding features)
    if method != 'decode':
        assert np.allclose(info.flatten().squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    info = neural_info(data.T, labels, axis=-1, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape[::-1]
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with string-valued labels
    groups = np.asarray(['cond1','cond2'])
    info = neural_info(data, groups[labels], axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output using groups argument to subset data
    info = neural_info(np.concatenate((data,data),axis=0), np.hstack((labels,labels+2)),
                       axis=0, method=method, groups=[2,3], **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with vector-valued data
    # For mutual info, computing bins over all data != over 1 channel,
    # so this ensures we have the same binning for both
    if method == 'mutual_info':
        bins = np.histogram_bin_edges(data, bins='fd')
        bins = np.stack((bins[:-1],bins[1:]),axis=1)
        extra_args['bins'] = bins
    info = neural_info(data[:,0], labels, axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert isinstance(info,float)
    # Skip decoding bc results are different when using only single channel/feature
    if method != 'decode': assert np.isclose(info, result[0], rtol=1e-2, atol=1e-2)
    if method == 'mutual_info': extra_args.pop('bins')

    if method == 'pev':
        # Test for consistent output using regression model (instead of ANOVA1)
        # Convert list of labels into a regression design matrix
        df = pd.DataFrame(labels,columns=['cond1'])
        model_formula= '1 + C(cond1, Sum)'
        design = dmatrix(model_formula,df)
        info = neural_info(data, design, axis=0, method=method, model='regress', **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == info_shape
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

        # Test for consistent output with constant column auto-appended in code
        df = pd.DataFrame(labels,columns=['cond1'])
        model_formula= 'C(cond1, Sum)'
        design = dmatrix(model_formula,df)
        info = neural_info(data, design, axis=0, method=method, model='regress', **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == info_shape
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

        # Test for consistent output with returning linear model stats (and check those)
        info, stats = neural_info(data, labels, axis=0, method=method, model='anova1',
                                  return_stats=True, **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == info_shape
        assert stats['F'].shape == info_shape
        assert stats['p'].shape == info_shape
        assert stats['mu'].shape == (2,n_chnls)
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)
        assert np.allclose(stats['F'].squeeze(), [11.5,12.21,50.04,31.91], rtol=1e-2, atol=1e-2)
        assert np.allclose(stats['p'].squeeze(), [3.25e-3,2.59e-3,1.35e-6,2.33e-5],
                           rtol=1e-5, atol=1e-5)
        mu = [[10.35,13.00, 3.74, 8.95], [19.45,19.15,20.77,19.00]]
        assert np.allclose(stats['mu'].squeeze(), mu, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        info = neural_info(data, labels, axis=0, method=method, foo=None, **extra_args)


@pytest.mark.parametrize('method, params, result',
                         [('pev',       {'omega':True},         (74.30,81.87,73.69,79.15)),
                          ('pev',       {'omega':False},        (76.67,83.58,76.11,81.10)),
                          ('decode',    {'decoder':'LDA'},      1.00),
                          ('decode',    {'decoder':'logistic'}, 0.80),
                          ('decode',    {'decoder':'SVM'},      0.95),
                          ('decode',    {'decoder':'custom'},   1.00)])
def test_one_way_info(one_way_data, method, params, result):
    """ Unit tests for info functions for computing info for one-way (multi-class) data """
    data, labels = one_way_data
    data_orig = data.copy()

    n = int(10)
    n_chnls = int(4)
    info_shape = (1,n_chnls)
    # Set any extra parameters for info function
    extra_args = params
    if method == 'decode':
        extra_args['seed'] = 1
        # Test for consistency when passing custom decoder function
        if extra_args['decoder'] == 'custom':
            extra_args['decoder'] = LinearDiscriminantAnalysis(priors=(1/3)*np.ones((3,)))

    # Basic test of shape, value of output
    info = neural_info(data, labels, axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':
        print(round(info,2))
        assert isinstance(info,float)
    else:
        print(np.round(info.squeeze(),2))
        assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for expected output with keepdims = False
    info = neural_info(data, labels, axis=0, method=method, keepdims=False, **extra_args)
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == (n_chnls,)
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)    

    # Test for consistent output with n-group form of neural computation function
    info = neural_info_ngroups(data[labels==0,:], data[labels==1,:], data[labels==2,:],
                               axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape (3rd axis)
    info = neural_info(data.reshape((n*3,int(n_chnls/2),int(n_chnls/2))), labels, axis=0,
                       method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    new_shape = (1,1,n_chnls/2) if method == 'decode' else (1,n_chnls/2,n_chnls/2)
    assert info.shape == new_shape
    # Skip decoding bc different results here due to different # of channels (decoding features)
    if method != 'decode':
        assert np.allclose(info.flatten().squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    info = neural_info(data.T, labels, axis=-1, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape[::-1]
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with string-valued labels
    groups = np.asarray(['cond1','cond2','cond3'])
    info = neural_info(data, groups[labels], axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output using groups argument to subset data
    info = neural_info(np.concatenate((data,data),axis=0), np.hstack((labels,labels+3)),
                       axis=0, method=method, groups=[3,4,5], **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    if method == 'decode':  assert isinstance(info,float)
    else:                   assert info.shape == info_shape
    assert np.allclose(np.asarray(info).squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with vector-valued data
    # For mutual info, computing bins over all data != over 1 channel,
    # so this ensures we have the same binning for both
    if method == 'mutual_info':
        bins = np.histogram_bin_edges(data, bins='fd')
        bins = np.stack((bins[:-1],bins[1:]),axis=1)
        extra_args ['bins'] = bins
    info = neural_info(data[:,0], labels, axis=0, method=method, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert isinstance(info,float)
    # Skip decoding bc results are different when using only single channel/feature
    if method != 'decode': assert np.isclose(info, result[0], rtol=1e-2, atol=1e-2)
    if method == 'mutual_info': extra_args.pop('bins')

    # For PEV method, test for consistent output using regression model (instead of ANOVA1)
    if method == 'pev':
        # Convert list of labels into a regression design matrix
        df = pd.DataFrame(labels,columns=['cond1'])
        model_formula= '1 + C(cond1, Sum)'
        design = dmatrix(model_formula,df)
        info = neural_info(data, design, axis=0, method=method, model='regress', **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == info_shape
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

        # Test for consistent output with constant column auto-appended in code
        df = pd.DataFrame(labels,columns=['cond1'])
        model_formula= 'C(cond1, Sum)'
        design = dmatrix(model_formula,df)
        info = neural_info(data, design, axis=0, method=method, model='regress', **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == info_shape
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

        # Test for consistent output with returning linear model stats (and check those)
        info, stats = neural_info(data, labels, axis=0, method=method, model='anova1',
                                  return_stats=True, **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == info_shape
        assert stats['F'].shape == info_shape
        assert stats['p'].shape == info_shape
        assert stats['mu'].shape == (3,n_chnls)
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)
        assert np.allclose(stats['F'].squeeze(), [44.36,68.72,43.00,57.93], rtol=1e-2, atol=1e-2)
        assert np.allclose(stats['p'].squeeze(), [2.93e-9,2.55e-11,4.04e-9,1.71e-10],
                           rtol=1e-11, atol=1e-11)
        mu = [[10.35,13.00,3.74,8.95], [19.45,19.15,20.77,19.00], [32.87,33.69,26.73,31.10]]
        assert np.allclose(stats['mu'].squeeze(), mu, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        info = neural_info(data, labels, axis=0, method=method, foo=None, **extra_args)


@pytest.mark.parametrize('method, interact, params, result',
                         [('pev', False,    {'omega':True},     (( 8.55, 2.91,27.23,15.67),
                                                                 (72.03,79.56,54.45,71.96))),
                          ('pev', False,    {'omega':False},    (( 9.07, 3.36,27.80,16.02),
                                                                 (72.84,80.33,55.14,72.48))),
                          ('pev', True,     {'omega':True},     (( 8.54, 2.93,27.23,15.65),
                                                                 (72.01,79.59,54.44,71.94),
                                                                 ( 0.05, 0.99, 0.31,-0.25))),
                          ('pev', True,     {'omega':False},    (( 9.07, 3.36,27.80,16.02),
                                                                 (72.84,80.33,55.14,72.48),
                                                                 ( 0.54, 1.41, 0.77, 0.07)))])
def test_two_way_info(two_way_data, method, interact, params, result):
    """ Unit tests for info func's for computing information for two-way (2-factor) data """
    data, labels = two_way_data
    data_orig = data.copy()
    labels = labels[:,0:2]
    result = np.asarray(result)
    # Set any extra parameters for info function
    extra_args = params

    n = int(10)
    n_chnls = int(4)
    n_terms = 2 + interact

    # Basic test of shape, value of output
    info = neural_info(data, labels, axis=0, method=method, model='anova2',
                       interact=interact, **extra_args)
    print(np.round(info,2))
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert info.shape == (n_terms,n_chnls)
    assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with different data array shape (3rd axis)
    info = neural_info(data.reshape((n*4,int(n_chnls/2),int(n_chnls/2))), labels, axis=0,
                       method=method, model='anova2', interact=interact, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert info.shape == (n_terms,n_chnls/2,n_chnls/2)
    assert np.allclose(info.flatten().squeeze(), result.flatten().squeeze(), rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed data dimensionality
    info = neural_info(data.T, labels, axis=-1, method=method, model='anova2',
                       interact=interact, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert info.shape == (n_chnls,n_terms)
    assert np.allclose(info.squeeze(), result.T, rtol=1e-2, atol=1e-2)

    # Test for consistent output with vector-valued data
    info = neural_info(data[:,0], labels, axis=0, method=method, model='anova2',
                       interact=interact, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert info.shape == (n_terms,)
    assert np.allclose(info, result[:,0], rtol=1e-2, atol=1e-2)

    # Test for consistent output with string-valued labels
    groups1 = np.asarray(['cond1.1','cond1.2'])
    groups2 = np.asarray(['cond2.1','cond2.2'])
    string_labels = np.stack((groups1[labels[:,0]],groups2[labels[:,1]]), axis=1)
    info = neural_info(data, string_labels, axis=0, method=method, model='anova2',
                       interact=interact, **extra_args)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert info.shape == (n_terms,n_chnls)
    assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

    # For PEV method, test for consistent output using regression model (instead of ANOVA1)
    if method == 'pev':
        # Convert list of labels into a regression design matrix
        df = pd.DataFrame(labels,columns=['cond1','cond2'])
        if interact:    model_formula = '1 + C(cond1, Sum)*C(cond2, Sum)'
        else:           model_formula = '1 + C(cond1, Sum) + C(cond2, Sum)'
        design = dmatrix(model_formula,df)
        info = neural_info(data, design, axis=0, method=method, model='regress', **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == (n_terms,n_chnls)
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

        # Test for consistent output with constant column auto-appended in code
        df = pd.DataFrame(labels,columns=['cond1','cond2'])
        if interact:    model_formula = 'C(cond1, Sum)*C(cond2, Sum)'
        else:           model_formula = 'C(cond1, Sum) + C(cond2, Sum)'
        design = dmatrix(model_formula,df)
        info = neural_info(data, design, axis=0, method=method, model='regress', **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == (n_terms,n_chnls)
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)

        # Test for consistent output with returning linear model stats (and check those)
        info, stats = neural_info(data, labels, axis=0, method=method, model='anova2',
                                  interact=interact, return_stats=True, **extra_args)
        assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
        assert info.shape == (n_terms,n_chnls)
        assert stats['F'].shape == (n_terms,n_chnls)
        assert stats['p'].shape == (n_terms,n_chnls)
        assert np.allclose(info.squeeze(), result, rtol=1e-2, atol=1e-2)
        if not interact:
            F = [[19.05, 7.82,61.94,52.94], [153.03,187.17,122.84,239.53]]
            p = [[9.44e-5,8.07e-3,1.69e-9,1.04e-8], [6.75e-15,2.92e-16,1.81e-13,5.40e-18]]
            mu = [[[21.61,23.34,15.24,20.02], [28.92,27.08,29.84,30.76]],
                  [[14.90,16.08,12.26,13.97], [35.63,34.34,32.82,36.81]]]
        else:
            F = [[18.60, 8.11,61.44,50.44], [149.45,194.08,121.84,228.22], [ 1.11, 3.40, 1.69, 0.21]]
            p = [[1.20e-4,7.24e-3,2.70e-9,2.41e-8],
                 [2.24e-14,4.52e-16,14.16e-13,3.70e-17],
                 [2.99e-1,7.33e-2,2.02e-1,6.52e-1]]
            mu = [[[21.61,23.34,15.24,20.02], [28.92,27.08,29.84,30.76]],
                  [[14.90,16.08,12.26,13.97], [35.63,34.34,32.82,36.81]],
                  [[10.35,13.00, 3.74,8.95], [19.45,19.15,20.77,19.00],
                   [32.87,33.69,26.73,31.10], [38.40,35.00,38.91,42.52]]]
        assert np.allclose(stats['F'].squeeze(), F, rtol=1e-2, atol=1e-2)
        assert np.allclose(stats['p'].squeeze(), p, rtol=1e-3, atol=1e-3)
        for term in range(len(mu)):
            print("MU", term, stats['mu'][term].shape)
            assert np.allclose(stats['mu'][term].squeeze(), mu[term], rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        info = neural_info(data, labels, axis=0, method=method, model='anova2', interact=interact,
                           foo=None)
        info = neural_info(data, design, axis=0, method=method, model='regress', foo=None)


def test_imports():
    """ Test different import methods for info module """
    # Import entire package
    import spynal
    spynal.info.dprime
    # Import module
    import spynal.info as info
    info.dprime
    # Import specific function from module
    from spynal.info import dprime
    dprime
